"""Module to render deodr scenes using OpenGL through pyrender.

Note that pyrender does not support camera distortion.
"""

import numpy as np

import pyrender

import trimesh


def arcsinc(x):
    if abs(x) > 1e-6:
        return np.arcsin(x) / x
    else:
        return 1


def min_rotation(vec1, vec2):
    """Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1,
                 aligns it with vec2.
    """
    a, b = (
        (vec1 / np.linalg.norm(vec1)).reshape(3),
        (vec2 / np.linalg.norm(vec2)).reshape(3),
    )
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    if s > 1e-6:
        d = (1 - c) / (s ** 2)
    else:
        d = 0.5
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * d
    return rotation_matrix


def deodr_directional_light_to_pyrender(deodr_directional_light):
    directional_light_intensity = np.linalg.norm(deodr_directional_light)
    if directional_light_intensity > 0:
        directional_light_direction = (
            deodr_directional_light / directional_light_intensity
        )
        directional_light_rotation = min_rotation(
            np.array([0, 0, -1]), directional_light_direction
        )
        pose = np.zeros((4, 4))
        pose[:3, :3] = directional_light_rotation
        pose[3, 3] = 1
        directional_light = pyrender.light.DirectionalLight(
            intensity=directional_light_intensity
        )
    else:
        directional_light = None
        pose = None

    return directional_light, pose


def deodr_mesh_to_pyrender(deodr_mesh):

    # trimesh and pyrender do to handle faces indices for texture
    # that are different from face indices for the 3d vertices
    # we need to duplicate vertices
    faces, mask_v, mask_vt = trimesh.visual.texture.unmerge_faces(
        deodr_mesh.faces, deodr_mesh.faces_uv
    )
    vertices = deodr_mesh.vertices[mask_v]
    deodr_mesh.compute_vertex_normals()
    vertex_normals = deodr_mesh.vertex_normals[mask_v]
    uv = deodr_mesh.uv[mask_vt]
    pyrender_uv = np.column_stack(
        (
            (uv[:, 0]) / deodr_mesh.texture.shape[0],
            (1 - uv[:, 1] / deodr_mesh.texture.shape[1]),
        )
    )

    material = None
    poses = None
    color_0 = None
    if material is None:
        base_color_texture = pyrender.texture.Texture(
            source=deodr_mesh.texture, source_channels="RGB"
        )
        material = pyrender.MetallicRoughnessMaterial(
            alphaMode="BLEND",
            baseColorFactor=[1, 1, 1, 1.0],
            metallicFactor=0,
            roughnessFactor=1,
            baseColorTexture=base_color_texture,
        )
        material.wireframe = False

    primitive = pyrender.Primitive(
        positions=vertices,
        normals=vertex_normals,
        texcoord_0=pyrender_uv,
        color_0=color_0,
        indices=faces,
        material=material,
        mode=pyrender.constants.GLTF.TRIANGLES,
        poses=poses,
    )

    return pyrender.Mesh(primitives=[primitive])


def render(deodr_scene, camera):
    """Render a deodr scene using pyrender"""
    bg_color = deodr_scene.background[0, 0]
    if not (np.all(deodr_scene.background == bg_color[None, None, :])):
        raise (
            BaseException(
                "pyrender does not support background image, please provide a"
                " backroundimage that correspond to a uniform color"
            )
        )
    pyrender_scene = pyrender.Scene(
        ambient_light=deodr_scene.light_ambient * np.ones((3)), bg_color=bg_color
    )

    if camera.distortion is not None:
        raise (BaseException("cameras with distortion not handled yet with pyrender"))

    # cam = pyrender.PerspectiveCamera(yfov=np.pi / 3.0, aspectRatio=1.414)
    cam = pyrender.IntrinsicsCamera(
        fx=camera.intrinsic[0, 0],
        fy=camera.intrinsic[1, 1],
        cx=camera.intrinsic[0, 2],
        cy=camera.intrinsic[1, 2],
    )

    # convert to pyrender directional light paramterization
    directional_light, directional_light_pose = deodr_directional_light_to_pyrender(
        deodr_scene.light_directional
    )
    if directional_light is not None:
        pyrender_scene.add(directional_light, pose=directional_light_pose)

    # convert the mesh
    pyrender_mesh = deodr_mesh_to_pyrender(deodr_scene.mesh)
    pyrender_scene.add(pyrender_mesh, pose=np.eye(4))

    # convert the camera
    m = camera.camera_to_world_mtx_4x4()
    pose_camera = m.dot(np.diag([1, -1, -1, 1]))

    pyrender_scene.add(cam, pose=pose_camera)

    # rendner
    width = camera.width
    height = camera.height
    r = pyrender.offscreen.OffscreenRenderer(width, height, point_size=1.0)
    image_pyrender, depth = r.render(pyrender_scene)
    return image_pyrender, depth



"""Example of fitting a hand mesh to a depth image."""

import datetime
import glob
import json
import os
import time

import cv2

import deodr
from deodr import ColoredTriMesh

from imageio import imsave

import matplotlib.pyplot as plt

import numpy as np


def run(
    dl_library="none",
    plot_curves=False,
    save_images=False,
    display=True,
    max_iter=300,
    n_subdivision=0,
):

    file_folder = os.path.dirname(__file__)

    if dl_library == "pytorch":
        from deodr.pytorch import MeshDepthFitter
    elif dl_library == "tensorflow":
        from deodr.tensorflow import MeshDepthFitter
    elif dl_library == "none":
        from deodr.mesh_fitter import MeshDepthFitter
    else:
        raise BaseException(f"unknown deep learning library {dl_library}")

    depth_image = np.fliplr(
        np.fromfile(os.path.join(deodr.data_path, "depth.bin"), dtype=np.float32)
        .reshape(240, 320)
        .astype(np.float)
    )
    depth_image = depth_image[20:-20, 60:-60]
    max_depth = 450
    depth_image[depth_image == 0] = max_depth
    depth_image = depth_image / max_depth

    obj_file = os.path.join(deodr.data_path, "hand.obj")
    faces, vertices = deodr.read_obj(obj_file)
    mesh = ColoredTriMesh(faces.copy(), vertices=vertices, nb_colors=0).subdivise(
        n_subdivision
    )

    euler_init = np.array([0.1, 0.1, 0.1])
    translation_init = np.zeros(3)

    hand_fitter = MeshDepthFitter(
        mesh.vertices, mesh.faces, euler_init, translation_init, cregu=1000
    )

    hand_fitter.set_image(depth_image, focal=241, distortion=[1, 0, 0, 0, 0])
    hand_fitter.set_max_depth(1)
    hand_fitter.set_depth_scale(110 / max_depth)
    energies = []
    durations = []
    start = time.time()

    iterfolder = os.path.join(file_folder, "./iterations/depth")
    if not os.path.exists(iterfolder):
        os.makedirs(iterfolder)

    for niter in range(max_iter):
        energy, synthetic_depth, diff_image = hand_fitter.step()
        energies.append(energy)
        durations.append(time.time() - start)
        if save_images or display:
            combined_image = np.column_stack(
                (depth_image, synthetic_depth, 3 * diff_image)
            )
            if display:
                cv2.imshow("animation", cv2.resize(combined_image, None, fx=2, fy=2))
            if save_images:
                imsave(
                    os.path.join(iterfolder, f"depth_hand_iter_{niter}.png"),
                    combined_image,
                )
        cv2.waitKey(1)

    with open(
        os.path.join(
            iterfolder,
            "depth_image_fitting_result_%s.json"
            % str(datetime.datetime.now()).replace(":", "_"),
        ),
        "w",
    ) as f:
        json.dump(
            {
                "label": f"{dl_library} {datetime.datetime.now()}",
                "durations": durations,
                "energies": energies,
            },
            f,
            indent=4,
        )

    if plot_curves:
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "depth_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(
                    json_data["durations"],
                    json_data["energies"],
                    label=json_data["label"],
                )
        plt.legend()
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "depth_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(json_data["energies"], label=json_data["label"])
        plt.legend()
        plt.show()

    return energies


def main():
    display = True
    n_subdivision = 0
    run(
        dl_library="none",
        plot_curves=False,
        save_images=False,
        display=display,
        n_subdivision=n_subdivision,
    )

    run(
        dl_library="pytorch",
        plot_curves=False,
        save_images=False,
        display=display,
        n_subdivision=n_subdivision,
    )

    run(
        dl_library="tensorflow",
        plot_curves=True,
        save_images=False,
        display=display,
        n_subdivision=n_subdivision,
    )


if __name__ == "__main__":
    main()



# distutils: language=c++
from libcpp cimport bool
cdef extern from "../C++/DifferentiableRenderer.h":
	ctypedef struct Scene:
		unsigned int* faces;
		unsigned int* faces_uv;
		double* depths
		double* uv
		double* ij
		double* shade
		double* colors
		bool* edgeflags
		bool* textured
		bool* shaded
		int     nb_triangles
		int nb_vertices;
		bool clockwise;
		bool backface_culling;
		int nb_uv;
		int     height
		int     width
		int     nb_colors
		double* texture
		int  texture_height
		int  texture_width
		double* background_image
		double* background_color
		double* uv_b
		double* ij_b
		double* shade_b
		double* colors_b
		double* texture_b
		bool strict_edge
		bool perspective_correct
		bool integer_pixel_centers
	void renderScene(Scene scene,double* image,double* z_buffer,double sigma,bool antialiase_error ,double* obs,double*  err_buffer)
	void renderScene_B(Scene scene,double* image,double* z_buffer,double* image_b,double sigma,bool antialiase_error ,double* obs,double*  err_buffer, double* err_buffer_b)



"""Example with fitting a colored hand mesh model to an image."""


import datetime
import glob
import json
import os
import time

import cv2

import deodr
from deodr import read_obj
from deodr import ColoredTriMesh

from imageio import imread, imsave

import matplotlib.pyplot as plt

import numpy as np


def run(
    dl_library="pytorch",
    plot_curves=True,
    save_images=True,
    display=True,
    max_iter=100,
    n_subdivision=0,
):
    if dl_library == "pytorch":
        from deodr.pytorch import MeshRGBFitterWithPose
    elif dl_library == "tensorflow":
        from deodr.tensorflow import MeshRGBFitterWithPose
    elif dl_library == "none":
        from deodr.mesh_fitter import MeshRGBFitterWithPose
    else:
        raise BaseException(f"unknown deep learning library {dl_library}")

    hand_image = (
        imread(os.path.join(deodr.data_path, "hand.png")).astype(np.double) / 255
    )
    obj_file = os.path.join(deodr.data_path, "hand.obj")
    faces, vertices = read_obj(obj_file)

    mesh = ColoredTriMesh(faces.copy(), vertices=vertices, nb_colors=3).subdivise(
        n_subdivision
    )

    default_color = np.array([0.4, 0.3, 0.25])
    default_light = {
        "directional": -np.array([0.1, 0.5, 0.4]),
        "ambient": np.array([0.6]),
    }

    euler_init = np.array([0, 0, 0])
    translation_init = np.mean(mesh.vertices, axis=0)
    # centering vertices
    mesh.vertices = mesh.vertices - translation_init[None, :]

    hand_fitter = MeshRGBFitterWithPose(
        mesh.vertices,
        mesh.faces,
        default_color=default_color,
        default_light=default_light,
        update_lights=True,
        update_color=True,
        euler_init=euler_init,
        translation_init=translation_init,
        cregu=1000,
    )

    hand_fitter.reset()

    background_color = np.median(
        np.row_stack(
            (
                hand_image[:10, :10, :].reshape(-1, 3),
                hand_image[-10:, :10, :].reshape(-1, 3),
                hand_image[-10:, -10:, :].reshape(-1, 3),
                hand_image[:10, -10:, :].reshape(-1, 3),
            )
        ),
        axis=0,
    )

    background_color = np.array([0.5, 0.6, 0.7])
    hand_fitter.set_image(hand_image, distortion=[-1, 0, 0, 0, 0])
    hand_fitter.set_background_color(background_color)
    energies = []
    durations = []
    start = time.time()

    iterfolder = "./iterations/rgb"
    if not os.path.exists(iterfolder):
        os.makedirs(iterfolder)

    for niter in range(max_iter):
        energy, image, diff_image = hand_fitter.step()
        energies.append(energy)
        durations.append(time.time() - start)
        if display or save_images:
            combined_image = np.column_stack(
                (hand_image, image, np.tile(diff_image[:, :, None], (1, 1, 3)))
            )
            if display:
                cv2.imshow(
                    "animation",
                    cv2.resize(combined_image[:, :, ::-1], None, fx=2, fy=2),
                )
            if save_images:
                imsave(
                    os.path.join(iterfolder, f"hand_iter_{niter}.png"), combined_image
                )
        cv2.waitKey(1)

    # save convergence curve
    with open(
        os.path.join(
            iterfolder,
            "rgb_image_fitting_result_%s.json"
            % str(datetime.datetime.now()).replace(":", "_"),
        ),
        "w",
    ) as f:
        json.dump(
            {
                "label": f"{dl_library} {datetime.datetime.now()}",
                "durations": durations,
                "energies": energies,
            },
            f,
            indent=4,
        )

    # compare with previous runs
    if plot_curves:
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(
                    json_data["durations"],
                    json_data["energies"],
                    label=json_data["label"],
                )
                plt.xlabel("duration in seconds")
                plt.ylabel("energies")
        plt.legend()
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(json_data["energies"], label=json_data["label"])
                plt.xlabel("interations")
                plt.ylabel("energies")
        plt.legend()
        plt.show()
    return energies


def main():

    display = True
    save_images = False
    n_subdivision = 1

    run(
        dl_library="pytorch",
        plot_curves=False,
        display=display,
        save_images=save_images,
        n_subdivision=n_subdivision,
    )

    run(
        dl_library="none",
        plot_curves=False,
        display=display,
        save_images=save_images,
        n_subdivision=n_subdivision,
    )
    run(
        dl_library="tensorflow",
        plot_curves=True,
        display=display,
        save_images=save_images,
         n_subdivision=n_subdivision,
    )


if __name__ == "__main__":
    main()



"""Pytorch interface to deodr."""

import numpy as np

import torch

from .. import differentiable_renderer_cython
from ..differentiable_renderer import Camera, Scene3D


class CameraPytorch(Camera):
    """Pytorch implementation of the camera class."""

    def __init__(self, extrinsic, intrinsic, height, width, distortion=None):
        super().__init__(
            extrinsic, intrinsic, height, width, distortion=distortion, checks=False
        )

    def world_to_camera(self, points_3d):
        assert isinstance(points_3d, torch.Tensor)
        return torch.cat(
            (points_3d, torch.ones((points_3d.shape[0], 1), dtype=torch.double)), dim=1
        ).mm(torch.tensor(self.extrinsic.T))

    def left_mul_intrinsic(self, projected):
        return torch.cat(
            (projected, torch.ones((projected.shape[0], 1), dtype=torch.double)), dim=1
        ).mm(torch.tensor(self.intrinsic[:2, :].T))

    def column_stack(self, values):
        return torch.stack(values, dim=1)


class TorchDifferentiableRenderer2DFunc(torch.autograd.Function):
    """Pytorch implementation of the 2D rendering function."""

    @staticmethod
    def forward(ctx, ij, colors, scene):
        nb_color_chanels = colors.shape[1]
        image = np.empty((scene.height, scene.width, nb_color_chanels))
        z_buffer = np.empty((scene.height, scene.width))
        ctx.scene = scene
        scene.ij = ij.detach().numpy()  # should automatically detached according to
        # https://pytorch.org/docs/master/notes/extending.html
        scene.colors = colors.detach().numpy()
        differentiable_renderer_cython.renderScene(scene, 1, image, z_buffer)
        ctx.save_for_backward(ij, colors)
        ctx.image = (
            image.copy()
        )  # making a copy to keep the antializaed image for visualization ,
        # could be optional
        ctx.z_buffer = z_buffer
        return torch.as_tensor(image)

    @staticmethod
    def backward(ctx, image_b):
        scene = ctx.scene
        scene.uv_b = np.zeros(scene.uv.shape)
        scene.ij_b = np.zeros(scene.ij.shape)
        scene.shade_b = np.zeros(scene.shade.shape)
        scene.colors_b = np.zeros(scene.colors.shape)
        scene.texture_b = np.zeros(scene.texture.shape)
        differentiable_renderer_cython.renderSceneB(
            scene, 1, ctx.image, ctx.z_buffer, image_b.numpy()
        )
        return torch.as_tensor(scene.ij_b), torch.as_tensor(scene.colors_b), None


TorchDifferentiableRender2D = TorchDifferentiableRenderer2DFunc.apply


class Scene3DPytorch(Scene3D):
    """Pytorch implementation of deodr 3D scenes."""

    def __init__(self):
        super().__init__()

    def set_light(self, light_directional, light_ambient):
        if not (isinstance(light_directional, torch.Tensor)):
            light_directional = torch.tensor(light_directional)
        self.light_directional = light_directional
        self.light_ambient = light_ambient

    def _compute_vertices_colors_with_illumination(self):
        vertices_luminosity = (
            torch.relu(
                -torch.sum(self.mesh.vertex_normals * self.light_directional, dim=1)
            )
            + self.light_ambient
        )
        return self.mesh.vertices_colors * vertices_luminosity[:, None]

    def _render_2d(self, ij, colors):
        self.depths = self.depths.detach()
        return TorchDifferentiableRender2D(ij, colors, self), self.depths



"""Pytorch implementation of a triangulated mesh."""


import numpy as np

import torch

from ..triangulated_mesh import TriMesh, TriMeshAdjacencies


def print_grad(name):
    def hook(grad):
        print(f"grad {name} = {grad}")

    return hook


class TriMeshAdjacenciesPytorch(TriMeshAdjacencies):
    """Class that stores adjacency matrices and methods that use this adjacencies using pytorch sparse matrices.
    Unlike the TriMesh class there are no vertices stored in this class.
    """

    def __init__(self, faces, clockwise=False):
        super().__init__(faces, clockwise)
        self.faces_torch = torch.LongTensor(faces)
        i = self.faces_torch.flatten()
        j = torch.LongTensor(
            np.tile(np.arange(self.nb_faces)[:, None], [1, 3]).flatten()
        )
        self._vertices_faces_torch = torch.sparse.DoubleTensor(
            torch.stack((i, j)),
            torch.ones((self.nb_faces, 3), dtype=torch.float64).flatten(),
            torch.Size((self.nb_vertices, self.nb_faces)),
        )

    def compute_face_normals(self, vertices):
        triangles = vertices[self.faces_torch, :]
        u = triangles[::, 1] - triangles[::, 0]
        v = triangles[::, 2] - triangles[::, 0]
        if self.clockwise:
            n = -torch.cross(u, v)
        else:
            n = torch.cross(u, v)
        l2 = (n ** 2).sum(dim=1)
        norm = l2.sqrt()
        nn = n / norm[:, None]
        return nn

    def compute_vertex_normals(self, face_normals):
        n = self._vertices_faces_torch.mm(face_normals)
        l2 = (n ** 2).sum(dim=1)
        norm = l2.sqrt()
        return n / norm[:, None]

    def edge_on_silhouette(self, vertices_2d):
        return super().edge_on_silhouette(vertices_2d.detach().numpy())


class TriMeshPytorch(TriMesh):
    """Pytorch implementation of a triangulated mesh."""

    def __init__(self, faces, vertices=None, clockwise=False):
        super().__init__(faces, vertices, clockwise)

    def compute_adjacencies(self):
        self.adjacencies = TriMeshAdjacenciesPytorch(self.faces)


class ColoredTriMeshPytorch(TriMeshPytorch):
    """Pytorch implementation of colored a triangulated mesh."""

    def __init__(
        self,
        faces,
        vertices=None,
        clockwise=False,
        faces_uv=None,
        uv=None,
        texture=None,
        colors=None,
    ):
        super(ColoredTriMeshPytorch, self).__init__(
            faces, vertices=vertices, clockwise=clockwise
        )
        self.faces_uv = faces_uv
        self.uv = uv
        self.texture = texture
        self.colors = colors
        self.textured = not (self.texture is None)

    def set_vertices_colors(self, colors):
        self.vertices_colors = colors



"""Modules containing classes to fit 3D meshes to images using differentiable rendering."""

import copy

import numpy as np

import scipy.sparse.linalg
import scipy.spatial.transform.rotation

from . import Camera, ColoredTriMesh, LaplacianRigidEnergy, Scene3D, TriMesh
from .tools import (
    normalize,
    normalize_backward,
    qrot,
    qrot_backward,
    check_jacobian_finite_differences,
)


class MeshDepthFitter:
    """Class to fit a deformable mesh to a depth image."""

    def __init__(
        self,
        vertices,
        faces,
        euler_init,
        translation_init,
        cregu=2000,
        inertia=0.96,
        damping=0.05
    ):
        self.cregu = cregu
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 1
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.1
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.mesh = TriMesh(
            faces, vertices=vertices
        )  # we do a copy to avoid negative stride not support by pytorch
        object_center = vertices.mean(axis=0)
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius

        self.scene = Scene3D()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = copy.copy(vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)

        self.reset()

    def set_mesh_transform_init(self, euler, translation):
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self):
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices_init.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

    def set_max_depth(self, max_depth):
        self.scene.max_depth = max_depth
        self.scene.set_background_color(np.array([max_depth], dtype=np.float))

    def set_depth_scale(self, depth_scale):
        self.depthScale = depth_scale

    def set_image(self, mesh_image, focal=None, distortion=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width
        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            height=self.height,
            width=self.width,
        )
        self.iter = 0

    def render(self):
        q_normalized = normalize(
            self.transform_quaternion
        )  # that will lead to a gradient that is in the tangeant space
        vertices_transformed = (
            qrot(q_normalized, self.vertices) + self.transform_translation
        )
        self.mesh.set_vertices(vertices_transformed)
        self.depth_not_clipped = self.scene.render_depth(
            self.camera, depth_scale=self.depthScale,
        )
        depth = np.clip(self.depth_not_clipped, 0, self.scene.max_depth)
        return depth

    def render_backward(self, depth_b):
        self.scene.clear_gradients()
        depth_b[self.depth_not_clipped < 0] = 0
        depth_b[self.depth_not_clipped > self.scene.max_depth] = 0
        self.scene.render_depth_backward(depth_b)
        vertices_transformed_b = self.scene.mesh.vertices_b
        self.transform_translation_b = np.sum(vertices_transformed_b, axis=0)
        q_normalized = normalize(self.transform_quaternion)
        q_normalized_b, self.vertices_b = qrot_backward(
            q_normalized, self.vertices, vertices_transformed_b
        )
        self.transform_quaternion_b = normalize_backward(
            self.transform_quaternion, q_normalized_b
        )  # that will lead to a gradient that is in the tangeant space
        return

    def step(self):

        self.vertices = self.vertices - np.mean(self.vertices, axis=0)[None, :]
        depth = self.render()

        diff_image = np.sum((depth - self.mesh_image[:, :, None]) ** 2, axis=2)
        energy_data = np.sum(diff_image)
        depth_b = 2 * (depth - self.mesh_image[:, :, None])
        self.render_backward(depth_b)

        self.vertices_b = self.vertices_b - np.mean(self.vertices_b, axis=0)[None, :]
        grad_data = self.vertices_b
        # update v

        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        def mult_and_clamp(x, a, t):
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia
        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * self.inertia + (1 - self.inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -self.transform_quaternion_b,
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation
        step_translation = mult_and_clamp(
            -self.transform_translation_b,
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation

        self.iter += 1
        return energy, depth[:, :, 0], diff_image


class MeshRGBFitterWithPose:
    """Class to fit a deformable mesh to a color image."""

    def __init__(
        self,
        vertices,
        faces,
        euler_init,
        translation_init,
        default_color,
        default_light,
        cregu=2000,
        inertia=0.96,
        damping=0.05,
        update_lights=True,
        update_color=True,
    ):
        self.cregu = cregu

        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light = default_light
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(faces.copy(), vertices=vertices, nb_colors=3)
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 9]) * object_radius

        self.scene = Scene3D()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = copy.copy(vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_background_color(self, background_color):
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(self, euler, translation):
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self):
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light["directional"])
        self.light_ambient = copy.copy(self.default_light["ambient"])

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = np.zeros(self.light_ambient.shape)
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_image(self, mesh_image, focal=None, distortion=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def render(self):
        q_normalized = normalize(
            self.transform_quaternion
        )  # that will lead to a gradient that is in the tangeant space
        vertices_transformed = (
            qrot(q_normalized, self.vertices) + self.transform_translation
        )
        self.mesh.set_vertices(vertices_transformed)
        self.scene.set_light(
            light_directional=self.light_directional, light_ambient=self.light_ambient
        )
        self.mesh.set_vertices_colors(
            np.tile(self.mesh_color, (self.mesh.nb_vertices, 1))
        )
        image = self.scene.render(self.camera)
        return image

    def render_backward(self, image_b):
        self.scene.clear_gradients()
        self.scene.render_backward(image_b)
        self.mesh_color_b = np.sum(self.mesh.vertices_colors_b, axis=0)
        self.light_directional_b = self.scene.light_directional_b
        self.light_ambient_b = self.scene.light_ambient_b
        vertices_transformed_b = self.scene.mesh.vertices_b
        self.transform_translation_b = np.sum(vertices_transformed_b, axis=0)
        q_normalized = normalize(self.transform_quaternion)
        q_normalized_b, self.vertices_b = qrot_backward(
            q_normalized, self.vertices, vertices_transformed_b
        )
        self.transform_quaternion_b = normalize_backward(
            self.transform_quaternion, q_normalized_b
        )  # that will lead to a gradient that is in the tangeant space
        return

    def step(self):
        self.vertices = self.vertices - np.mean(self.vertices, axis=0)[None, :]

        image = self.render()

        diff_image = np.sum((image - self.mesh_image) ** 2, axis=2)
        image_b = 2 * (image - self.mesh_image)
        energy_data = np.sum(diff_image)

        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        self.render_backward(image_b)

        self.vertices_b = self.vertices_b - np.mean(self.vertices_b, axis=0)[None, :]
        # update v
        grad = self.vertices_b + grad_rigidity

        def mult_and_clamp(x, a, t):
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -self.transform_quaternion_b,
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation
        step_translation = mult_and_clamp(
            -self.transform_translation_b,
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -self.light_directional_b * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step = -self.light_ambient_b * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient
        # update mesh color
        step = -self.mesh_color_b * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image, diff_image


class MeshRGBFitterWithPoseMultiFrame:
    """Class to fit a deformable mesh to multiple color images."""

    def __init__(
        self,
        vertices,
        faces,
        euler_init,
        translation_init,
        default_color,
        default_light,
        cregu=2000,
        cdata=1,
        inertia=0.97,
        damping=0.15,
        update_lights=True,
        update_color=True,
    ):
        self.cregu = cregu
        self.cdata = cdata
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00005
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00004
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light = default_light
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(faces, vertices, nb_colors=3)
        object_center = vertices.mean(axis=0)
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 6]) * object_radius

        self.scene = Scene3D()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = copy.copy(vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_background_color(self, background_color):
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(self, euler, translation):
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self):
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light["directional"])
        self.light_ambient = copy.copy(self.default_light["ambient"])

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = np.zeros(self.light_ambient.shape)
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_images(self, mesh_images, focal=None):
        self.width = mesh_images[0].shape[1]
        self.height = mesh_images[0].shape[0]
        assert mesh_images[0].ndim == 3
        self.mesh_images = mesh_images
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def set_image(self, mesh_image, focal=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def render(self, idframe=None):
        unormalized_quaternion = self.transform_quaternion[idframe]
        q_normalized = normalize(
            unormalized_quaternion
        )  # that will lead to a gradient that is in the tangeant space
        vertices_transformed = (
            qrot(q_normalized, self.vertices) + self.transform_translation[idframe]
        )
        self.mesh.set_vertices(vertices_transformed)
        self.scene.set_light(
            light_directional=self.light_directional, light_ambient=self.light_ambient
        )
        self.mesh.set_vertices_colors(
            np.tile(self.mesh_color, (self.mesh.nb_vertices, 1))
        )
        image = self.scene.render(self.camera)
        self.store_backward["render"] = (idframe, unormalized_quaternion, q_normalized)
        return image

    def clear_gradients(self):
        self.light_directional_b = np.zeros(self.light_directional.shape)
        self.light_ambient_b = np.zeros(self.light_ambient.shape)
        self.vertices_b = np.zeros(self.vertices.shape)
        self.transform_quaternion_b = np.zeros(self.transform_quaternion.shape)
        self.transform_translation_b = np.zeros(self.transform_translation.shape)
        self.mesh_color_b = np.zeros(self.mesh_color.shape)
        self.store_backward = {}

    def render_backward(self, image_b):
        idframe, unormalized_quaternion, q_normalized = self.store_backward["render"]
        self.scene.clear_gradients()
        self.scene.render_backward(image_b)
        self.mesh_color_b += np.sum(self.mesh.vertices_colors_b, axis=0)
        self.light_directional_b += self.scene.light_directional_b
        self.light_ambient_b += self.scene.light_ambient_b
        vertices_transformed_b = self.scene.mesh.vertices_b
        self.transform_translation_b[idframe] += np.sum(vertices_transformed_b, axis=0)
        q_normalized_b, vertices_b = qrot_backward(
            q_normalized, self.vertices, vertices_transformed_b
        )
        self.vertices_b += vertices_b
        self.transform_quaternion_b[idframe] += normalize_backward(
            unormalized_quaternion, q_normalized_b
        )  # that will lead to a gradient that is in the tangeant space
        return

    def energy_data(self, vertices, return_images=True):
        self.vertices = vertices
        image = [None] * self.nb_facesrames
        diff_image = [None] * self.nb_facesrames
        image_b = [None] * self.nb_facesrames
        energy_datas = [None] * self.nb_facesrames
        self.clear_gradients()
        coef_data = self.cdata / self.nb_facesrames
        for idframe in range(self.nb_facesrames):
            image[idframe] = self.render(idframe=idframe)
            diff_image[idframe] = np.sum(
                (image[idframe] - self.mesh_images[idframe]) ** 2, axis=2
            )
            image_b = coef_data * 2 * (image[idframe] - self.mesh_images[idframe])
            energy_datas[idframe] = coef_data * np.sum(diff_image[idframe])
            self.render_backward(image_b)
        energy_data = np.sum(energy_datas)
        if return_images:
            return energy_data, image, diff_image
        else:
            return energy_data

    def step(self, check_gradient=False):

        self.vertices = self.vertices - np.mean(self.vertices, axis=0)[None, :]

        self.nb_facesrames = len(self.mesh_images)

        energy_data, image, diff_image = self.energy_data(self.vertices)
        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices)

        if check_gradient:

            def func(x):
                return self.rigid_energy.evaluate(
                    x, return_grad=False, return_hessian=False
                )

            check_jacobian_finite_differences(
                grad_rigidity.flatten(), func, self.vertices
            )

            def func(x):
                return self.energy_data(x, return_images=False)

            grad_data = self.vertices_b.copy()
            check_jacobian_finite_differences(grad_data.flatten(), func, self.vertices)

        energy = energy_data + energy_rigid
        print(
            f"iter {self.iter} Energy={energy} : EData={energy_data} E_rigid={energy_rigid}"
        )

        if self.iter < 500:
            self.vertices_b = (
                self.vertices_b - np.mean(self.vertices_b, axis=0)[None, :]
            )
        # update v
        grad = self.vertices_b + grad_rigidity

        def mult_and_clamp(x, a, t):
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )

        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation

        step_quaternion = mult_and_clamp(
            -self.transform_quaternion_b,
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation
        step_translation = mult_and_clamp(
            -self.transform_translation_b,
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -self.light_directional_b * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step = -self.light_ambient_b * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient
        # update mesh color
        step = -self.mesh_color_b * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image, diff_image



"""Modules containing pytorch classes to fit 3D meshes to images using differentiable rendering."""


import copy

import numpy as np

import scipy.sparse.linalg
import scipy.spatial.transform.rotation

import torch

from . import CameraPytorch, LaplacianRigidEnergyPytorch, Scene3DPytorch
from .triangulated_mesh_pytorch import ColoredTriMeshPytorch as ColoredTriMesh
from .triangulated_mesh_pytorch import TriMeshPytorch as TriMesh
from .. import LaplacianRigidEnergy


def print_grad(name):
    # to visualize the gradient of a variable use
    # variable_name.register_hook(print_grad('variable_name'))
    def hook(grad):
        print(f"grad {name} = {grad}")

    return hook


def qrot(q, v):
    qr = q[None, :].repeat(v.shape[0], 1)
    qvec = qr[:, :-1]
    uv = torch.cross(qvec, v, dim=1)
    uuv = torch.cross(qvec, uv, dim=1)
    return v + 2 * (qr[:, [3]] * uv + uuv)


class MeshDepthFitterEnergy(torch.nn.Module):
    """Pytorch module to fit a deformable mesh to a depth image."""

    def __init__(self, vertices, faces, euler_init, translation_init, cregu=2000):
        super(MeshDepthFitterEnergy, self).__init__()
        self.mesh = TriMesh(
            faces[:, ::-1].copy(), vertices
        )  # we do a copy to avoid negative stride not supported by pytorch
        object_center = vertices.mean(axis=0)
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius
        self.scene = Scene3DPytorch()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergyPytorch(self.mesh, vertices, cregu)
        self.Vinit = copy.copy(self.mesh.vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler_init
        ).as_quat()
        self.transform_translation_init = translation_init
        self._vertices = torch.nn.Parameter(
            torch.tensor(self.Vinit, dtype=torch.float64)
        )
        self.quaternion = torch.nn.Parameter(
            torch.tensor(self.transform_quaternion_init, dtype=torch.float64)
        )
        self.translation = torch.nn.Parameter(
            torch.tensor(self.transform_translation_init, dtype=torch.float64)
        )

    def set_max_depth(self, max_depth):
        self.scene.max_depth = max_depth
        self.scene.set_background_color(np.array([max_depth], dtype=np.float))

    def set_depth_scale(self, depth_scale):
        self.depthScale = depth_scale

    def set_image(self, mesh_image, focal=None, distortion=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        t = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, t))
        self.camera = CameraPytorch(
            extrinsic=extrinsic, intrinsic=intrinsic, distortion=distortion
        )
        self.iter = 0

    def forward(self):
        q_normalized = self.quaternion / self.quaternion.norm()
        print(self.quaternion.norm())
        vertices_centered = self._vertices - torch.mean(self._vertices, dim=0)[None, :]
        v_transformed = qrot(q_normalized, vertices_centered) + self.translation
        self.mesh.set_vertices(v_transformed)
        depth_scale = 1 * self.depthScale
        depth = self.scene.render_depth(
            self.CameraMatrix,
            width=self.width,
            height=self.height,
            depth_scale=depth_scale,
        )
        depth = torch.clamp(depth, 0, self.scene.max_depth)
        diff_image = torch.sum(
            (depth - torch.tensor(self.mesh_image[:, :, None])) ** 2, dim=2
        )
        self.depth = depth
        self.diff_image = diff_image
        energy_data = torch.sum(diff_image)
        energy_rigid = self.rigid_energy.evaluate(
            self._vertices, return_grad=False, return_hessian=False
        )
        energy = energy_data + energy_rigid
        self.loss = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))
        return self.loss


class MeshDepthFitterPytorchOptim:
    """Pytorch optimizer to fit a deformable mesh to an image."""

    def __init__(
        self, vertices, faces, euler_init, translation_init, cregu=2000, lr=0.8
    ):
        self.energy = MeshDepthFitterEnergy(
            vertices, faces, euler_init, translation_init, cregu
        )
        params = self.energy.parameters()
        self.optimizer = torch.optim.LBFGS(params, lr=0.8, max_iter=1)
        # self.optimizer = torch.optim.SGD(params, lr=0.000005, momentum=0.1,
        # dampening=0.1        )
        # self.optimizer =torch.optim.RMSprop(params, lr=1e-3, alpha=0.99,  eps=1e-8,
        # weight_decay=0,  momentum=0.001)
        # self.optimizer = torch.optim.Adadelta(params, lr=0.1, rho=0.95,
        #   eps=1e-6,  weight_decay=0)
        # self.optimizer = torch.optim.Adagrad(self.energy.parameters(), lr=0.02)

    def set_image(self, depth_image, focal):
        self.energy.set_image(depth_image, focal=focal)

    def set_max_depth(self, max_depth):
        self.energy.set_max_depth(max_depth)

    def set_depth_scale(self, depth_scale):
        self.energy.set_depth_scale(depth_scale)

    def step(self):
        def closure():
            self.optimizer.zero_grad()
            loss = self.energy()
            loss.backward()
            return loss

        self.optimizer.step(closure)
        # self.iter += 1
        return (
            self.energy.loss,
            self.energy.Depth[:, :, 0].detach().numpy(),
            self.energy.diffImage.detach().numpy(),
        )


class MeshDepthFitter:
    """Class to fit a deformable mesh to a depth image."""

    def __init__(
        self,
        vertices,
        faces,
        euler_init,
        translation_init,
        cregu=2000,
        inertia=0.96,
        damping=0.05,
    ):
        self.cregu = cregu
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.1
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.mesh = TriMesh(
            faces.copy()
        )  # we do a copy to avoid negative stride not support by pytorch
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius

        self.scene = Scene3DPytorch()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = torch.tensor(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_mesh_transform_init(self, euler, translation):
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self):
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices_init.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

    def set_max_depth(self, max_depth):
        self.scene.max_depth = max_depth
        self.scene.set_background_color([max_depth])

    def set_depth_scale(self, depth_scale):
        self.depthScale = depth_scale

    def set_image(self, mesh_image, focal=None, distortion=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraPytorch(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )
        self.iter = 0

    def step(self):
        self.vertices = self.vertices - torch.mean(self.vertices, dim=0)[None, :]
        # vertices_with_grad = self.vertices.clone().requires_grad(True)
        vertices_with_grad = self.vertices.clone().detach().requires_grad_(True)
        vertices_with_grad_centered = (
            vertices_with_grad - torch.mean(vertices_with_grad, dim=0)[None, :]
        )
        quaternion_with_grad = torch.tensor(
            self.transform_quaternion, dtype=torch.float64, requires_grad=True
        )
        translation_with_grad = torch.tensor(
            self.transform_translation, dtype=torch.float64, requires_grad=True
        )

        q_normalized = (
            quaternion_with_grad / quaternion_with_grad.norm()
        )  # that will lead to a gradient that is in the tangent space
        vertices_with_grad_transformed = (
            qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
        )

        self.mesh.set_vertices(vertices_with_grad_transformed)

        depth_scale = 1 * self.depthScale
        depth = self.scene.render_depth(self.camera, depth_scale=depth_scale)
        depth = torch.clamp(depth, 0, self.scene.max_depth)

        diff_image = torch.sum(
            (depth - torch.tensor(self.mesh_image[:, :, None])) ** 2, dim=2
        )
        loss = torch.sum(diff_image)

        loss.backward()
        energy_data = loss.detach().numpy()

        grad_data = vertices_with_grad.grad.numpy()

        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices.numpy())
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        def mult_and_clamp(x, a, t):
            return np.minimum(np.maximum(x * a, -t), t)

        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * self.inertia + (1 - self.inertia) * step_vertices
        )
        self.vertices = self.vertices + torch.tensor(self.speed_vertices)
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_with_grad.grad.numpy(),
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * self.inertia + (1 - self.inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation

        step_translation = mult_and_clamp(
            -translation_with_grad.grad.numpy(),
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * self.inertia
            + (1 - self.inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation

        self.iter += 1
        return energy, depth[:, :, 0].detach().numpy(), diff_image.detach().numpy()


class MeshRGBFitterWithPose:
    """Class to fit a deformable mesh to a color image."""

    def __init__(
        self,
        vertices,
        faces,
        euler_init,
        translation_init,
        default_color,
        default_light,
        cregu=2000,
        inertia=0.96,
        damping=0.05,
        update_lights=True,
        update_color=True,
    ):
        self.cregu = cregu

        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light = default_light
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(
            faces.copy()
        )  # we do a copy to avoid negative stride not support by pytorch
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 9]) * object_radius

        self.scene = Scene3DPytorch()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergyPytorch(self.mesh, vertices, cregu)
        self.vertices_init = torch.tensor(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_background_color(self, background_color):
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(self, euler, translation):
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self):
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light["directional"])
        self.light_ambient = copy.copy(self.default_light["ambient"])

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = np.zeros(self.light_ambient.shape)
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_image(self, mesh_image, focal=None, distortion=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraPytorch(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )
        self.iter = 0

    def step(self):
        self.vertices = self.vertices - torch.mean(self.vertices, dim=0)[None, :]
        vertices_with_grad = self.vertices.clone().detach().requires_grad_(True)
        vertices_with_grad_centered = (
            vertices_with_grad - torch.mean(vertices_with_grad, dim=0)[None, :]
        )
        quaternion_with_grad = torch.tensor(
            self.transform_quaternion, dtype=torch.float64, requires_grad=True
        )
        translation_with_grad = torch.tensor(
            self.transform_translation, dtype=torch.float64, requires_grad=True
        )

        light_directional_with_grad = torch.tensor(
            self.light_directional, dtype=torch.float64, requires_grad=True
        )
        light_ambient_with_grad = torch.tensor(
            self.light_ambient, dtype=torch.float64, requires_grad=True
        )
        mesh_color_with_grad = torch.tensor(
            self.mesh_color, dtype=torch.float64, requires_grad=True
        )

        q_normalized = (
            quaternion_with_grad / quaternion_with_grad.norm()
        )  # that will lead to a gradient that is in the tangent space
        vertices_with_grad_transformed = (
            qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
        )
        self.mesh.set_vertices(vertices_with_grad_transformed)

        self.scene.set_light(
            light_directional=light_directional_with_grad,
            light_ambient=light_ambient_with_grad,
        )
        self.mesh.set_vertices_colors(
            mesh_color_with_grad.repeat([self.mesh.nb_vertices, 1])
        )

        image = self.scene.render(self.camera)

        diff_image = torch.sum((image - torch.tensor(self.mesh_image)) ** 2, dim=2)
        loss = torch.sum(diff_image)

        loss.backward()
        energy_data = loss.detach().numpy()

        grad_data = vertices_with_grad.grad

        (
            energy_rigid,
            grad_rigidity,
            approx_hessian_rigidity,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid.numpy()
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        def mult_and_clamp(x, a, t):
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad.numpy(), self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + torch.tensor(self.speed_vertices)
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_with_grad.grad.numpy(),
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )

        # update translation
        step_translation = mult_and_clamp(
            -translation_with_grad.grad.numpy(),
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -light_directional_with_grad.grad.numpy() * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step = -light_ambient_with_grad.grad.numpy() * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient
        # update mesh color
        step = -mesh_color_with_grad.grad.numpy() * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image.detach().numpy(), diff_image.detach().numpy()



"""Tensorflow interface to deodr."""

import numpy as np

import tensorflow as tf

from .. import differentiable_renderer_cython
from ..differentiable_renderer import Camera, Scene3D


class CameraTensorflow(Camera):
    """Tensorflow implementation of the camera class."""

    def __init__(self, extrinsic, intrinsic, height, width, distortion=None):
        super().__init__(
            extrinsic, intrinsic, height, width, distortion=distortion, checks=False
        )

    def world_to_camera(self, points_3d):
        assert isinstance(points_3d, tf.Tensor)
        return tf.linalg.matmul(
            tf.concat(
                (points_3d, tf.ones((points_3d.shape[0], 1), dtype=points_3d.dtype)),
                axis=1,
            ),
            tf.constant(self.extrinsic.T),
        )

    def left_mul_intrinsic(self, projected):
        assert isinstance(projected, tf.Tensor)
        return tf.linalg.matmul(
            tf.concat(
                (projected, tf.ones((projected.shape[0], 1), dtype=projected.dtype)),
                axis=1,
            ),
            tf.constant(self.intrinsic[:2, :].T),
        )

    def column_stack(self, values):
        return tf.stack(values, axis=1)


def TensorflowDifferentiableRender2D(ij, colors, scene):
    """Tensorflow implementation of the 2D rendering function."""

    @tf.custom_gradient
    def forward(
        ij, colors
    ):  # using inner function as we don't differentate w.r.t scene
        nb_color_chanels = colors.shape[1]
        image = np.empty((scene.height, scene.width, nb_color_chanels))
        z_buffer = np.empty((scene.height, scene.width))
        scene.ij = np.array(ij)  # should automatically detached according to
        # https://pytorch.org/docs/master/notes/extending.html
        scene.colors = np.array(colors)
        scene.depths = np.array(scene.depths)
        differentiable_renderer_cython.renderScene(scene, 1, image, z_buffer)

        def backward(image_b):
            scene.uv_b = np.zeros(scene.uv.shape)
            scene.ij_b = np.zeros(scene.ij.shape)
            scene.shade_b = np.zeros(scene.shade.shape)
            scene.colors_b = np.zeros(scene.colors.shape)
            scene.texture_b = np.zeros(scene.texture.shape)
            image_copy = (
                image.copy()
            )  # making a copy to avoid removing antialiasing on the image returned by
            # the forward pass (the c++ backpropagation undo antialiasing), could be
            # optional if we don't care about getting aliased images
            differentiable_renderer_cython.renderSceneB(
                scene, 1, image_copy, z_buffer, image_b.numpy()
            )
            return tf.constant(scene.ij_b), tf.constant(scene.colors_b)

        return tf.convert_to_tensor(image), backward

    return forward(ij, colors)


class Scene3DTensorflow(Scene3D):
    """Tensorflow implementation of deodr 3D scenes."""

    def __init__(self):
        super().__init__()

    def set_light(self, light_directional, light_ambient):
        if not (isinstance(light_directional, tf.Tensor)):
            light_directional = tf.constant(light_directional)
        self.light_directional = light_directional
        self.light_ambient = light_ambient

    def _compute_vertices_colors_with_illumination(self):
        vertices_luminosity = (
            tf.nn.relu(
                -tf.reduce_sum(
                    self.mesh.vertex_normals * self.light_directional, axis=1
                )
            )
            + self.light_ambient
        )
        return self.mesh.vertices_colors * vertices_luminosity[:, None]

    def _render_2d(self, ij, colors):

        return TensorflowDifferentiableRender2D(ij, colors, self), self.depths



"""Tensorflow implementation of a triangulated mesh."""

import tensorflow as tf

from .tools import scipy_sparse_matrix_to_tensorflow
from ..triangulated_mesh import TriMesh, TriMeshAdjacencies


class TriMeshAdjacenciesTensorflow(TriMeshAdjacencies):
    """Class that stores adjacency matrices and methods that use this adjacencies using tensorflow sparse matrices.
    Unlike the TriMesh class there are no vertices stored in this class.
    """

    def __init__(self, faces):
        super().__init__(faces)
        self.faces_tf = tf.constant(faces)
        self._vertices_faces_tf = scipy_sparse_matrix_to_tensorflow(
            self._vertices_faces
        )

    def compute_face_normals(self, vertices):
        tris = tf.gather(vertices, self.faces)
        u = tris[::, 1] - tris[::, 0]
        v = tris[::, 2] - tris[::, 0]
        if self.clockwise:
            n = -tf.linalg.cross(u, v)
        else:
            n = tf.linalg.cross(u, v)
        norm = tf.sqrt(tf.reduce_sum(n ** 2, axis=1))
        return n / norm[:, None]

    def compute_vertex_normals(self, face_normals):
        n = tf.sparse.sparse_dense_matmul(self._vertices_faces_tf, face_normals)
        norm = tf.sqrt(tf.reduce_sum(n ** 2, axis=1))
        return n / norm[:, None]

    def edge_on_silhouette(self, vertices_2d):
        return super().edge_on_silhouette(vertices_2d.numpy())


class TriMeshTensorflow(TriMesh):
    """Tensorflow implementation of a triangulated mesh."""

    def __init__(self, faces, vertices=None, clockwise=False):
        super().__init__(faces, vertices, clockwise)

    def compute_adjacencies(self):
        self.adjacencies = TriMeshAdjacenciesTensorflow(self.faces)


class ColoredTriMeshTensorflow(TriMeshTensorflow):
    """Tensorflow implementation of a colored triangulated mesh."""

    def __init__(
        self,
        faces,
        vertices=None,
        clockwise=False,
        faces_uv=None,
        uv=None,
        texture=None,
        colors=None,
    ):
        super(ColoredTriMeshTensorflow, self).__init__(
            faces, vertices=vertices, clockwise=clockwise
        )
        self.faces_uv = faces_uv
        self.uv = uv
        self.texture = texture
        self.colors = colors
        self.textured = not (self.texture is None)

    def set_vertices_colors(self, colors):
        self.vertices_colors = colors



"""Pytorch implementation of an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

import numpy as np

import torch

from ..laplacian_rigid_energy import LaplacianRigidEnergy


def scipy_sparse_to_torch(sparse_matrix):
    coo = sparse_matrix.tocoo()
    values = coo.data
    indices = np.vstack((coo.row, coo.col))
    i = torch.LongTensor(indices.astype(np.int32))
    v = torch.DoubleTensor(values)
    shape = coo.shape
    return torch.sparse.DoubleTensor(i, v, torch.Size(shape))


class LaplacianRigidEnergyPytorch(LaplacianRigidEnergy):
    """Pytorch class that implements an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

    def __init__(self, mesh, vertices, cregu):
        super().__init__(mesh, vertices, cregu)
        self.cT_torch = scipy_sparse_to_torch(self.cT)

    def evaluate(
        self, vertices, return_grad=True, return_hessian=True, refresh_rotations=True
    ):
        assert isinstance(vertices, torch.Tensor)
        if vertices.requires_grad:
            diff = (vertices - self.vertices_ref).flatten()
            grad_vertices = self.cregu * (
                self.cT_torch.matmul(diff[:, None])
            ).reshape_as(vertices)
            energy = 0.5 * diff.dot(grad_vertices.flatten())
            return energy
        else:
            diff = (vertices - torch.tensor(self.vertices_ref)).flatten()
            # gradV = self.cregu*(self.cT_torch.matmul(diff[:,None])).reshape_as(V)
            # 40x slower than scipy !
            grad_vertices = torch.tensor(
                self.cregu * (self.cT * (diff[:, None].numpy())).reshape(vertices.shape)
            )
            energy = 0.5 * diff.dot(grad_vertices.flatten())
            if not (return_grad):
                assert not (return_hessian)
                return energy
            if not (return_hessian):
                return energy, grad_vertices
            return energy, grad_vertices, self.approx_hessian



"""Tensorflow implementation of an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

import tensorflow as tf

from .tools import scipy_sparse_matrix_to_tensorflow
from ..laplacian_rigid_energy import LaplacianRigidEnergy


class LaplacianRigidEnergyTensorflow(LaplacianRigidEnergy):
    """Tensorflow class that implements an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

    def __init__(self, mesh, vertices, cregu):
        super().__init__(mesh, vertices, cregu)
        self.cT_tf = scipy_sparse_matrix_to_tensorflow(self.cT)

    def evaluate(self, vertices, return_grad=True, return_hessian=True):
        assert isinstance(vertices, tf.Tensor)
        diff = tf.reshape(vertices - tf.constant(self.vertices_ref), [-1])
        grad_vertices = tf.reshape(
            self.cregu * tf.sparse.sparse_dense_matmul(self.cT_tf, diff[:, None]),
            vertices.shape,
        )  # 5 times slower than scipy
        energy = 0.5 * tf.reduce_sum(diff * tf.reshape(grad_vertices, [-1]))
        if not (return_grad):
            assert not (return_hessian)
            return energy
        if not (return_hessian):
            return energy, grad_vertices
        return energy, grad_vertices, self.approx_hessian



"""Example with fitting a 32 triangles soup to an image."""


import copy
import os

import cv2

import deodr
from deodr import differentiable_renderer_cython
from deodr.differentiable_renderer import Scene2D

from imageio import imread

import matplotlib.pyplot as plt

import numpy as np


def create_example_scene(n_tri=30, width=200, height=200, clockwise=False):

    material = np.double(imread(os.path.join(deodr.data_path, "trefle.jpg"))) / 255
    height_material = material.shape[0]
    width_material = material.shape[1]

    scale_matrix = np.array([[height, 0], [0, width]])
    scale_material = np.array([[height_material - 1, 0], [0, width_material - 1]])

    triangles = []
    for _ in range(n_tri):

        tmp = scale_matrix.dot(
            np.random.rand(2, 1).dot(np.ones((1, 3)))
            + 0.5 * (-0.5 + np.random.rand(2, 3))
        )
        while np.abs(np.linalg.det(np.vstack((tmp, np.ones((3)))))) < 1500:
            tmp = scale_matrix.dot(
                np.random.rand(2, 1).dot(np.ones((1, 3)))
                + 0.5 * (-0.5 + np.random.rand(2, 3))
            )

        if np.linalg.det(np.vstack((tmp, np.ones((3))))) > 0:
            tmp = np.fliplr(tmp)
        triangle = {}
        triangle["ij"] = tmp.T
        triangle["depths"] = np.random.rand(1) * np.ones(
            (3, 1)
        )  # constant depth triangles to avoid collisions
        triangle["textured"] = np.random.rand(1) > 0.5

        if triangle["textured"]:
            triangle["uv"] = (
                scale_material.dot(np.array([[0, 1, 0.2], [0, 0.2, 1]])).T + 1
            )  # texture coordinate of the vertices
            triangle["shade"] = np.random.rand(3, 1)  # shade  intensity at each vertex
            triangle["colors"] = np.zeros((3, 3))
            triangle["shaded"] = True
        else:
            triangle["uv"] = np.zeros((3, 2))
            triangle["shade"] = np.zeros((3, 1))
            triangle["colors"] = np.random.rand(3, 3)
            # colors of the vertices (can be gray, rgb color,or even other dimension
            # vectors) when using simple linear interpolation across triangles
            triangle["shaded"] = False

        triangle["edgeflags"] = np.array(
            [True, True, True]
        )  # all edges are discontinuity edges as no triangle pair share an edge
        triangles.append(triangle)

    scene = {}
    for key in triangles[0].keys():
        scene[key] = np.squeeze(
            np.vstack([np.array(triangle[key]) for triangle in triangles])
        )
    scene["faces"] = np.arange(3 * n_tri).reshape(-1, 3).astype(np.uint32)
    scene["faces_uv"] = np.arange(3 * n_tri).reshape(-1, 3).astype(np.uint32)

    if clockwise:
        scene["faces"] = np.fliplr(scene["faces"])
        scene["faces_uv"] = np.fliplr(scene["faces_uv"])

    scene["clockwise"] = clockwise
    scene["height"] = height
    scene["width"] = width
    scene["texture"] = material
    scene["nb_colors"] = 3
    scene["background_color"] = None
    scene["background_image"] = np.tile(
        np.array([0.3, 0.5, 0.7])[None, None, :], (height, width, 1)
    )
    scene["perspective_correct"] = False
    scene["backface_culling"] = True
    return Scene2D(**scene)


def run(nb_max_iter=500, display=True, clockwise=False):
    print("process id=%d" % os.getpid())

    np.random.seed(2)
    scene_gt = create_example_scene(clockwise=clockwise)
    antialiase_error = False
    sigma = 1

    image_target = np.zeros((scene_gt.height, scene_gt.width, scene_gt.nb_colors))

    z_buffer = np.zeros((scene_gt.height, scene_gt.width))
    differentiable_renderer_cython.renderScene(scene_gt, sigma, image_target, z_buffer)

    n_vertices = len(scene_gt.depths)

    displacement_magnitude_ij = 10
    displacement_magnitude_uv = 0
    displacement_magnitude_colors = 0

    alpha_ij = 0.01
    beta_ij = 0.80
    alpha_uv = 0.03
    beta_uv = 0.80
    alpha_color = 0.001
    beta_color = 0.70

    max_uv = np.array(scene_gt.texture.shape[:2]) - 1

    scene_init = copy.deepcopy(scene_gt)
    scene_init.ij = (
        scene_gt.ij + np.random.randn(n_vertices, 2) * displacement_magnitude_ij
    )
    scene_init.uv = (
        scene_gt.uv + np.random.randn(n_vertices, 2) * displacement_magnitude_uv
    )
    scene_init.uv = np.maximum(scene_init.uv, 0)
    scene_init.uv = np.minimum(scene_init.uv, max_uv)
    scene_init.colors = (
        scene_gt.colors + np.random.randn(n_vertices, 3) * displacement_magnitude_colors
    )

    final_loss = {}

    for antialiase_error in [True, False]:
        np.random.seed(2)
        scene_iter = copy.deepcopy(scene_init)

        speed_ij = np.zeros((n_vertices, 2))
        speed_uv = np.zeros((n_vertices, 2))
        speed_color = np.zeros((n_vertices, 3))

        losses = []
        for niter in range(nb_max_iter):
            image, depth, loss_image, loss = scene_iter.render_compare_and_backward(
                sigma, antialiase_error, image_target
            )
            print(f"iter {niter} loss = {loss}")
            # imsave(os.path.join(iterfolder,f'soup_{niter}.png'), combinedIMage)

            losses.append(loss)
            if loss_image.ndim == 2:
                loss_image = np.broadcast_to(loss_image[:, :, None], image.shape)
            if display:
                cv2.waitKey(1)
                cv2.imshow(
                    "animation",
                    np.column_stack((image_target, image, loss_image))[:, :, ::-1],
                )

            if displacement_magnitude_ij > 0:
                speed_ij = beta_ij * speed_ij - scene_iter.ij_b * alpha_ij
                scene_iter.ij = scene_iter.ij + speed_ij

            if displacement_magnitude_colors > 0:
                speed_color = (
                    beta_color * speed_color - scene_iter.colors_b * alpha_color
                )
                scene_iter.colors = scene_iter.colors + speed_color

            if displacement_magnitude_uv > 0:
                speed_uv = beta_uv * speed_uv - scene_iter.uv_b * alpha_uv
                scene_iter.uv = scene_iter.uv + speed_uv
                scene_iter.uv = max(scene_iter.uv, 0)
                scene_iter.uv = min(scene_iter.uv, max_uv)
        if display:
            plt.plot(losses, label="antialiaseError=%d" % antialiase_error)
        final_loss[antialiase_error] = loss
    if display:
        plt.legend()
        plt.show()
    return final_loss


if __name__ == "__main__":
    run()



"""example fitting a texture mesh to an image using face images"""

import cv2

from deodr.differentiable_renderer import Scene2D

import matplotlib.pyplot as plt

import numpy as np

from scipy.spatial import Delaunay

import sklearn.datasets
from sklearn import decomposition


def main():
    faces = sklearn.datasets.fetch_olivetti_faces()

    # plt.imshow(faces.images[0])

    faces_pca = decomposition.PCA(n_components=150, whiten=True)
    faces_pca.fit(faces.data)
    plt.imshow(faces_pca.mean_.reshape(faces.images[0].shape), cmap=plt.cm.bone)

    # coefs = faces_pca.transform(faces.data)
    # print(faces)

    # faces_pca.inverse_transform(coefs[[0]])

    # create a regular grid and a triangulation  of the 2D image
    N = 5
    points = np.column_stack(
        [t.flatten() for t in np.meshgrid(np.arange(N + 1) / N, np.arange(N + 1) / N)]
    )
    tri = Delaunay(points)
    triangles = tri.simplices.astype(np.uint32)
    plt.triplot(points[:, 0], points[:, 1], triangles)
    on_border = np.any((points == 0) | (points == 1), axis=1)

    # deform the grid
    max_displacement = 0.5
    np.random.seed(0)
    points_deformed_gt = (
        points + (np.random.rand(*points.shape) - 0.5) * max_displacement / N
    )
    points_deformed_gt[on_border] = points[on_border]
    plt.triplot(points_deformed_gt[:, 0], points_deformed_gt[:, 1], triangles)
    # plt.show()
    nb_points = points.shape[0]
    nb_triangles = triangles.shape[0]
    ij = points_deformed_gt * 64 - 0.5
    uv = points * 64 + 0.5
    textured = np.ones((nb_triangles), dtype=np.bool)
    shaded = np.ones((nb_triangles), dtype=np.bool)
    depths = np.ones((nb_points))
    shade = np.ones((nb_points))
    colors = np.ones((nb_points, 1))
    edgeflags = np.zeros((nb_triangles, 3), dtype=np.bool)
    height = 64
    width = 64
    nb_colors = 1
    texture = faces.images[10][:, :, None]
    background_color = np.array([0])

    scene_gt = Scene2D(
        triangles,
        triangles,
        ij,
        depths,
        textured,
        uv,
        shade,
        colors,
        shaded,
        edgeflags,
        height,
        width,
        nb_colors,
        texture,
        background_image=None,
        background_color=background_color,
        clockwise=True,
        backface_culling=True,
    )

    image_gt, _ = scene_gt.render(sigma=1)
    plt.subplot(3, 1, 1)
    plt.imshow(np.squeeze(texture, axis=2))
    plt.subplot(3, 1, 2)
    plt.imshow(np.squeeze(image_gt, axis=2))
    plt.subplot(3, 1, 3)
    plt.imshow(np.squeeze(np.abs(image_gt - texture), axis=2))

    # plt.show()
    np.max(texture - image_gt)

    scene = Scene2D(
        triangles,
        triangles,
        ij,
        depths,
        textured,
        uv,
        shade,
        colors,
        shaded,
        edgeflags,
        height,
        width,
        nb_colors,
        texture,
        background_image=None,
        background_color=background_color,
        clockwise=True,
        backface_culling=True,
    )
    # cv2.namedWindow('animation',cv2.WINDOW_NORMAL)
    # cv2.resizeWindow('animation', 600,600)

    rescale_factor = 10

    def fun(points_deformed, pca_coefs):
        ij = points_deformed * 64 - 0.5
        # face=faces_pca.inverse_transform(coefs[:,None])
        face = (faces_pca.mean_ + pca_coefs.dot(faces_pca.components_)).reshape(
            (64, 64)
        )
        scene.ij = ij
        scene.texture = face[:, :, None]
        print("render")
        image, z_buffer, diff_image, err = scene.render_compare_and_backward(
            obs=image_gt, sigma=1
        )
        print("np.max(np.abs(scene.ij_b))=%f" % np.max(np.abs(scene.ij_b)))
        print("E=%f" % err)
        print("done")
        image_gt_zoomed = cv2.resize(
            (image_gt.copy() * 255).astype(np.uint8),
            None,
            fx=rescale_factor,
            fy=rescale_factor,
            interpolation=cv2.INTER_NEAREST,
        )
        image_zoomed = cv2.resize(
            (image.copy() * 255).astype(np.uint8),
            None,
            fx=rescale_factor,
            fy=rescale_factor,
            interpolation=cv2.INTER_NEAREST,
        )
        diff_image_zoomed = cv2.resize(
            (np.abs(diff_image) * 255 * 10).astype(np.uint8),
            None,
            fx=rescale_factor,
            fy=rescale_factor,
            interpolation=cv2.INTER_NEAREST,
        )

        cv2.polylines(
            image_gt_zoomed,
            (points_deformed_gt[tri.simplices] * 64 * rescale_factor).astype(np.int32),
            isClosed=True,
            color=(0, 0, 0),
            lineType=cv2.LINE_AA,
        )

        cv2.polylines(
            image_zoomed,
            (points_deformed[tri.simplices] * 64 * rescale_factor).astype(np.int32),
            isClosed=True,
            color=(0, 0, 0),
            lineType=cv2.LINE_AA,
        )
        cv2.imshow(
            "animation",
            np.column_stack((image_gt_zoomed, image_zoomed, diff_image_zoomed)),
        )
        cv2.waitKey(1)

        # get gradient on pca coefs
        coefs_grad = faces_pca.components_.dot(scene.texture_b.flatten())
        points_deformed_grad = scene.ij_b * 64
        print(np.max(np.abs(points_deformed_grad)))
        grads = {"points_deformed": points_deformed_grad, "pca_coefs": coefs_grad}
        return err, grads

    nb_iter = 100

    pca_coefs = np.zeros((faces_pca.n_components))
    points_deformed = points.copy()

    variables = {"points_deformed": points_deformed, "pca_coefs": pca_coefs}
    lambdas = {"points_deformed": 0.0001, "pca_coefs": 0.5}

    for niter in range(nb_iter):

        E, grads = fun(**variables)
        print(f"iter{niter} E={E}")
        for name in variables.keys():
            variables[name] = variables[name] - lambdas[name] * grads[name]

        variables["points_deformed"][on_border] = points[on_border]

    cv2.waitKey()


if __name__ == "__main__":
    main()



"""Implementation of an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

import copy

import scipy
import scipy.sparse


class LaplacianRigidEnergy:
    """Class that implements an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

    def __init__(self, mesh, vertices, cregu):
        self.cT = scipy.sparse.kron(
            mesh.adjacencies.laplacian.T * mesh.adjacencies.laplacian,
            scipy.sparse.eye(3),
        ).tocsr()
        self.vertices_ref = copy.copy(vertices)
        self.mesh = mesh
        self.cregu = cregu
        self.approx_hessian = self.cregu * self.cT
        n_components, labels = scipy.sparse.csgraph.connected_components(
            csgraph=self.mesh.adjacencies.adjacency_vertices,
            directed=False,
            return_labels=True,
        )
        if n_components > 1:
            raise (
                BaseException(
                    "You have more than one connected component in your mesh."
                )
            )

    def evaluate(
        self, vertices, return_grad=True, return_hessian=True, refresh_rotations=True
    ):

        diff = (vertices - self.vertices_ref).flatten()
        grad_vertices = self.cregu * (self.cT * diff).reshape((vertices.shape[0], 3))
        energy = 0.5 * diff.dot(grad_vertices.flatten())
        if not (return_grad):
            assert not (return_hessian)
            return energy
        if not (return_hessian):
            return energy, grad_vertices

        return energy, grad_vertices, self.approx_hessian



"""Setup script for the DEODR project."""

import os
import re
from setuptools import setup, find_packages

from Cython.Build import cythonize

import numpy as np


# compilation mode for debuging
# extensions = [
# Extension("differentiable_renderer_cython",
# ["DEODR/differentiable_renderer_cython.pyx"]
# ,extra_compile_args=["-Zi", "/Od"]
# ,extra_link_args=["-debug"],
# undef_macros = [ "NDEBUG" ]
# )
# ]

extensions = "deodr/differentiable_renderer_cython.pyx"

my_modules = cythonize(extensions, annotate=True, language="c++")

libname = "deodr"

with open(os.path.join(os.path.dirname(__file__), "deodr", "__init__.py")) as fp:
    for line in fp:
        m = re.search(r'^\s*__version__\s*=\s*([\'"])([^\'"]+)\1\s*$', line)
        if m:
            version = m.group(2)
            break
    else:
        raise RuntimeError("Unable to find own __version__ string")
print(f"version = {version}")
setup(
    name=libname,
    version=version,
    author="Martin de La Gorce",
    author_email="martin.delagorce@gmail.com",
    description="A differentiable renderer with Pytorch,Tensorflow and Matlab interfaces.",
    url="https://github.com/martinResearch/DEODR",
    license="BSD",
    packages=find_packages(),
    package_data={"deodr": ["*.pyx", "*.pxd", "data/*.*", "data/**/*.*"]},
    data_files=[("C++", ["C++/DifferentiableRenderer.h"])],
    ext_modules=my_modules,  # additional source file(s)),
    include_dirs=[np.get_include()],
    setup_requires=["numpy", "scipy", "cython"],
    install_requires=["numpy", "scipy"],
)



"""Modules containing Tensorflow classes to fit 3D meshes to images using differentiable rendering."""


import copy

import numpy as np

import scipy.sparse.linalg
import scipy.spatial.transform.rotation

import tensorflow as tf

from . import (
    CameraTensorflow,
    LaplacianRigidEnergyTensorflow,
    Scene3DTensorflow,
)
from .triangulated_mesh_tensorflow import ColoredTriMeshTensorflow as ColoredTriMesh
from .triangulated_mesh_tensorflow import TriMeshTensorflow as TriMesh
from .. import LaplacianRigidEnergy


def qrot(q, v):
    qr = tf.tile(q[None, :], (v.shape[0], 1))
    qvec = qr[:, :-1]
    uv = tf.linalg.cross(qvec, v)
    uuv = tf.linalg.cross(qvec, uv)
    return v + 2 * (qr[:, 3][None, 0] * uv + uuv)


class MeshDepthFitter:
    """Class to fit a deformable mesh to a depth image using Tensorflow."""

    def __init__(
        self,
        vertices,
        faces,
        euler_init,
        translation_init,
        cregu=2000,
        inertia=0.96,
        damping=0.05,
    ):
        self.cregu = cregu
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.1
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.mesh = TriMesh(
            faces, vertices
        )  # we do a copy to avoid negative stride not support by Tensorflow
        object_center = vertices.mean(axis=0)
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius

        self.scene = Scene3DTensorflow()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = tf.constant(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_mesh_transform_init(self, euler, translation):
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self):
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices_init.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

    def set_max_depth(self, max_depth):
        self.scene.max_depth = max_depth
        self.scene.set_background_color([max_depth])

    def set_depth_scale(self, depth_scale):
        self.depthScale = depth_scale

    def set_image(self, mesh_image, focal=None, distortion=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraTensorflow(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )
        self.iter = 0

    def step(self):
        self.vertices = (
            self.vertices - tf.reduce_mean(self.vertices, axis=0)[None, :]
        )  # centervertices because we have another paramter to control translations

        with tf.GradientTape() as tape:

            vertices_with_grad = tf.constant(self.vertices)
            quaternion_with_grad = tf.constant(self.transform_quaternion)
            translation_with_grad = tf.constant(self.transform_translation)

            tape.watch(vertices_with_grad)
            tape.watch(quaternion_with_grad)
            tape.watch(translation_with_grad)

            vertices_with_grad_centered = (
                vertices_with_grad - tf.reduce_mean(vertices_with_grad, axis=0)[None, :]
            )

            q_normalized = quaternion_with_grad / tf.norm(
                quaternion_with_grad
            )  # that will lead to a gradient that is in the tangeant space
            vertices_with_grad_transformed = (
                qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
            )

            self.mesh.set_vertices(vertices_with_grad_transformed)

            depth_scale = 1 * self.depthScale
            depth = self.scene.render_depth(self.camera, depth_scale=depth_scale,)
            depth = tf.clip_by_value(depth, 0, self.scene.max_depth)

            diff_image = tf.reduce_sum(
                (depth - tf.constant(self.mesh_image[:, :, None])) ** 2, axis=2
            )
            loss = tf.reduce_sum(diff_image)

            trainable_variables = [
                vertices_with_grad,
                quaternion_with_grad,
                translation_with_grad,
            ]
            vertices_grad, quaternion_grad, translation_grad = tape.gradient(
                loss, trainable_variables
            )

        energy_data = loss.numpy()

        grad_data = vertices_grad

        (
            energy_rigid,
            grad_rigidity,
            approx_hessian_rigidity,
        ) = self.rigid_energy.evaluate(self.vertices.numpy())
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        def mult_and_clamp(x, a, t):
            return np.minimum(np.maximum(x * a, -t), t)

        # update vertices
        step_vertices = mult_and_clamp(
            -grad.numpy(), self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * self.inertia + (1 - self.inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_grad.numpy(),
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * self.inertia + (1 - self.inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        # update translation
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        step_translation = mult_and_clamp(
            -translation_grad.numpy(),
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * self.inertia
            + (1 - self.inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation

        self.iter += 1
        return energy, depth[:, :, 0].numpy(), diff_image.numpy()


class MeshRGBFitterWithPose:
    """Class to fit a deformable mesh to a color image using Tensorflow."""

    def __init__(
        self,
        vertices,
        faces,
        euler_init,
        translation_init,
        default_color,
        default_light,
        cregu=2000,
        inertia=0.96,
        damping=0.05,
        update_lights=True,
        update_color=True,
    ):
        self.cregu = cregu

        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light = default_light
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(
            faces.copy()
        )  # we do a copy to avoid negative stride not support by Tensorflow
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 9]) * object_radius

        self.scene = Scene3DTensorflow()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergyTensorflow(self.mesh, vertices, cregu)
        self.vertices_init = tf.constant(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_background_color(self, background_color):
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(self, euler, translation):
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self):
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light["directional"])
        self.light_ambient = copy.copy(self.default_light["ambient"])

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = np.zeros(self.light_ambient.shape)
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_image(self, mesh_image, focal=None, distortion=None):
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraTensorflow(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def step(self):

        with tf.GradientTape() as tape:

            vertices_with_grad = tf.constant(self.vertices)
            quaternion_with_grad = tf.constant(self.transform_quaternion)
            translation_with_grad = tf.constant(self.transform_translation)

            light_directional_with_grad = tf.constant(self.light_directional)
            light_ambient_with_grad = tf.constant(self.light_ambient)
            mesh_color_with_grad = tf.constant(self.mesh_color)

            tape.watch(vertices_with_grad)
            tape.watch(quaternion_with_grad)
            tape.watch(translation_with_grad)

            tape.watch(light_directional_with_grad)
            tape.watch(light_ambient_with_grad)
            tape.watch(mesh_color_with_grad)

            vertices_with_grad_centered = (
                vertices_with_grad - tf.reduce_mean(vertices_with_grad, axis=0)[None, :]
            )

            q_normalized = quaternion_with_grad / tf.norm(
                quaternion_with_grad
            )  # that will lead to a gradient that is in the tangeant space
            vertices_with_grad_transformed = (
                qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
            )
            self.mesh.set_vertices(vertices_with_grad_transformed)

            self.scene.set_light(
                light_directional=light_directional_with_grad,
                light_ambient=light_ambient_with_grad,
            )
            self.mesh.set_vertices_colors(
                tf.tile(mesh_color_with_grad[None, :], [self.mesh.nb_vertices, 1])
            )

            image = self.scene.render(self.camera)

            diff_image = tf.reduce_sum(
                (image - tf.constant(self.mesh_image)) ** 2, axis=2
            )
            loss = tf.reduce_sum(diff_image)

            trainable_variables = [
                vertices_with_grad,
                quaternion_with_grad,
                translation_with_grad,
                light_directional_with_grad,
                light_ambient_with_grad,
                mesh_color_with_grad,
            ]
            (
                vertices_grad,
                quaternion_grad,
                translation_grad,
                light_directional_grad,
                light_ambient_grad,
                mesh_color_grad,
            ) = tape.gradient(loss, trainable_variables)

        energy_data = loss.numpy()

        grad_data = vertices_grad

        (
            energy_rigid,
            grad_rigidity,
            approx_hessian_rigidity,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid.numpy()
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        def mult_and_clamp(x, a, t):
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad.numpy(), self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_grad, self.step_factor_quaternion, self.step_max_quaternion
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        # update translation
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        step_translation = mult_and_clamp(
            -translation_grad, self.step_factor_translation, self.step_max_translation
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -light_directional_grad * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step = -light_ambient_grad * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient
        # update mesh color
        step = -mesh_color_grad * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image.numpy(), diff_image.numpy()



"""Small tools for tensorflow."""

import numpy as np

import tensorflow as tf


def scipy_sparse_matrix_to_tensorflow(x):
    coo = x.tocoo()
    indices = np.mat([coo.row, coo.col]).transpose()
    return tf.SparseTensor(indices, coo.data, coo.shape)



"""Example with fitting a colored hand mesh model to multiple images."""


import datetime
import glob
import json
import os
import time

import cv2

import deodr
from deodr import read_obj
from deodr.mesh_fitter import MeshRGBFitterWithPoseMultiFrame

from imageio import imread, imsave

import matplotlib.pyplot as plt

import numpy as np


def run(
    dl_library="pytorch",
    plot_curves=False,
    save_images=False,
    display=True,
    max_iter=400,
):

    file_folder = os.path.dirname(__file__)
    hand_images = [
        imread(file).astype(np.double) / 255
        for file in glob.glob(os.path.join(deodr.data_path, "./hand_multiview/*.jpg"))
    ]
    nb_frames = len(hand_images)

    obj_file = os.path.join(deodr.data_path, "hand.obj")
    faces, vertices = read_obj(obj_file)

    default_color = np.array([0.4, 0.3, 0.25]) * 1.5
    default_light = {
        "directional": -np.array([0.1, 0.5, 0.4]),
        "ambient": np.array([0.6]),
    }
    # default_light = {'directional':np.array([0.0,0.0,0.0]),'ambient':np.array([0.6])}

    euler_init = np.row_stack(
        [np.array([0, yrot, 0]) for yrot in np.linspace(-0.5, 0.5, 3)]
    )

    vertices = vertices - np.mean(vertices, axis=0)
    t_init = np.array([0, -0.2, 0.2])
    translation_init = np.tile(t_init[None, :], [nb_frames, 1])
    # centering vertices

    hand_fitter = MeshRGBFitterWithPoseMultiFrame(
        vertices,
        faces,
        default_color=default_color,
        default_light=default_light,
        update_lights=True,
        update_color=True,
        euler_init=euler_init,
        translation_init=translation_init,
        cregu=2000,
    )
    #  handFitter = MeshRGBFitter(vertices,faces,default_color,default_light,
    # update_lights =  True, update_color= True,cregu=1000)

    hand_fitter.reset()

    max_iter = 300

    hand_image = hand_images[0]
    background_color = np.median(
        np.row_stack(
            (
                hand_image[:10, :10, :].reshape(-1, 3),
                hand_image[-10:, :10, :].reshape(-1, 3),
                hand_image[-10:, -10:, :].reshape(-1, 3),
                hand_image[:10, -10:, :].reshape(-1, 3),
            )
        ),
        axis=0,
    )
    background_color = np.array([0, 0, 0])
    hand_fitter.set_images(hand_images)
    hand_fitter.set_background_color(background_color)
    energies = []
    durations = []
    start = time.time()

    iterfolder = os.path.join(file_folder, "./iterations/depth")
    if not os.path.exists(iterfolder):
        os.makedirs(iterfolder)

    for niter in range(max_iter):

        energy, image, diff_image = hand_fitter.step(check_gradient=False)
        energies.append(energy)
        durations.append(time.time() - start)
        if display or save_images:
            combined_image = np.column_stack(
                (
                    np.row_stack(hand_images),
                    np.row_stack(image),
                    np.tile(
                        np.row_stack(np.minimum(diff_image, 1))[:, :, None], (1, 1, 3)
                    ),
                )
            )
            if display:
                cv2.imshow(
                    "animation",
                    cv2.resize(combined_image[:, :, ::-1], None, fx=1, fy=1),
                )
            if save_images:
                imsave(
                    os.path.join(iterfolder, f"hand_iter_{niter}.png"),
                    (combined_image * 255).astype(np.uint8),
                )
        cv2.waitKey(1)

    # save convergence curve
    with open(
        os.path.join(
            iterfolder,
            "rgb_image_fitting_result_%s.json"
            % str(datetime.datetime.now()).replace(":", "_"),
        ),
        "w",
    ) as f:
        json.dump(
            {
                "label": f"{dl_library} {datetime.datetime.now()}",
                "durations": durations,
                "energies": energies,
            },
            f,
            indent=4,
        )

    # compare with previous runs
    if plot_curves:
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(
                    json_data["durations"],
                    json_data["energies"],
                    label=json_data["label"],
                )
                plt.xlabel("duration in seconds")
                plt.ylabel("energies")
        plt.legend()
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(json_data["energies"], label=json_data["label"])
                plt.xlabel("interations")
                plt.ylabel("energies")
        plt.legend()
        plt.show()


if __name__ == "__main__":
    display = True
    save_images = False
    run(dl_library="none", plot_curves=True, display=display, save_images=save_images)



"""Example of interactive 3D mesh visualization using DEODR and OpenCV."""

import argparse
import os
import time
import pickle

import cv2

import deodr
from deodr import differentiable_renderer
from deodr.triangulated_mesh import ColoredTriMesh

import matplotlib.pyplot as plt

import numpy as np

from scipy.spatial.transform import Rotation

import trimesh


class Interactor:
    """Class that implements various mouse interaction with the 3D scene."""

    def __init__(
        self,
        camera,
        mode="object_centered_trackball",
        object_center=None,
        rotation_speed=0.003,
        z_translation_speed=0.05,
        xy_translation_speed=0.01,
    ):
        self.left_is_down = False
        self.right_is_down = False
        self.middle_is_down = False
        self.mode = mode
        self.object_center = object_center
        self.rotation_speed = rotation_speed
        self.z_translation_speed = z_translation_speed
        self.xy_translation_speed = xy_translation_speed
        self.camera = camera

    def toggle_mode(self):
        if self.mode == "object_centered_trackball":
            self.mode = "camera_centered"
        else:
            self.mode = "object_centered_trackball"
        print(f"trackball mode = {self.mode}")

    def rotate(
        self,
        rot_vec,
    ):
        rotation = Rotation.from_rotvec(np.array(rot_vec))
        if self.mode == "camera_centered":
            self.camera.extrinsic = rotation.as_matrix().dot(self.camera.extrinsic)
        else:
            n_rotation = rotation.as_matrix().dot(self.camera.extrinsic[:, :3])
            nt = (
                self.camera.extrinsic[:, :3].dot(self.object_center)
                + self.camera.extrinsic[:, 3]
                - n_rotation.dot(self.object_center)
            )
            self.camera.extrinsic = np.column_stack((n_rotation, nt))

    def mouse_callback(self, event, x, y, flags, param):
        if event == 0 and flags == 0:
            return
        if event == cv2.EVENT_LBUTTONDOWN:
            self.left_is_down = True
            self.x_last = x
            self.y_last = y
        # check to see if the left mouse button was released
        if event == cv2.EVENT_LBUTTONUP:
            self.left_is_down = False
        if event == cv2.EVENT_RBUTTONDOWN:
            self.right_is_down = True
            self.x_last = x
            self.y_last = y
            # check to see if the left mouse button was released
        if event == cv2.EVENT_RBUTTONUP:
            self.right_is_down = False

        if event == cv2.EVENT_MBUTTONDOWN:
            self.middle_is_down = True
            self.x_last = x
            self.y_last = y
        # check to see if the left mouse button was released
        if event == cv2.EVENT_MBUTTONUP:
            self.middle_is_down = False

        self.ctrl_is_down = flags & cv2.EVENT_FLAG_CTRLKEY
        self.shift_is_down = flags & cv2.EVENT_FLAG_SHIFTKEY

        if self.left_is_down and not (self.ctrl_is_down):

            if self.mode == "camera_centered":
                rot_vec = [
                    -0.3 * self.rotation_speed * (y - self.y_last),
                    0.3 * self.rotation_speed * (x - self.x_last),
                    0,
                ]
                self.rotate(rot_vec)

                # assert np.allclose(center_in_camera, self.camera.world_to_camera(self.object_center))
                self.x_last = x
                self.y_last = y

            elif self.mode == "object_centered_trackball":

                self.rotate(
                    [
                        self.rotation_speed * (y - self.y_last),
                        -self.rotation_speed * (x - self.x_last),
                        0,
                    ]
                )

                self.x_last = x
                self.y_last = y
            else:
                raise (BaseException(f"unknown camera mode {self.mode}"))

        if self.right_is_down and self.shift_is_down:
            delta_y = self.y_last - y
            ratio = np.power(2, delta_y / 20)
            self.camera.intrinsic[0, 0] = self.camera.intrinsic[0, 0] * ratio
            self.camera.intrinsic[1, 1] = self.camera.intrinsic[1, 1] * ratio
            self.x_last = x
            self.y_last = y

        if self.right_is_down and not (self.ctrl_is_down):
            if self.mode in ["camera_centered", "object_centered_trackball"]:
                if np.abs(self.y_last - y) >= np.abs(self.x_last - x):
                    self.camera.extrinsic[2, 3] += self.z_translation_speed * (
                        self.y_last - y
                    )
                else:
                    self.rotate(
                        [
                            0,
                            0,
                            -self.rotation_speed * (self.x_last - x),
                        ]
                    )
                self.x_last = x
                self.y_last = y

            else:
                raise (BaseException(f"unknown camera mode {self.mode}"))

        if self.middle_is_down or (self.left_is_down and self.ctrl_is_down):
            # translation

            object_depth = (
                self.camera.extrinsic[2, :3].dot(self.object_center)
                + self.camera.extrinsic[2, 3]
            )

            tx = self.xy_translation_speed * object_depth * (x - self.x_last)
            ty = self.xy_translation_speed * object_depth * (y - self.y_last)

            self.object_center -= (
                self.camera.extrinsic[0, :3] * tx + self.camera.extrinsic[1, :3] * ty
            )
            self.camera.extrinsic[0, 3] += tx
            self.camera.extrinsic[1, 3] += ty
            self.x_last = x
            self.y_last = y

    def print_help(self):
        help_str = ""
        help_str += "Mouse:\n"
        if self.mode == "object_centered_trackball":

            help_str += (
                "mouse left + vertical motion: rotate object along camera x axis\n"
            )
            help_str += (
                "mouse left + horizontal motion: rotate object along camera y axis\n"
            )
            help_str += (
                "mouse right + vertical motion: translate object along camera z axis\n"
            )
            help_str += (
                "mouse right + horizontal motion: rotate object along camera z axis\n"
            )
            help_str += "CTRL + mouse left + vertical motion: translate object along camera y axis\n"
            help_str += "CTRL + mouse left + horizontal motion: translate object along camera x axis\n"

            help_str += "SHIFT + mouse left + vertical motion: change the camera field of view\n"
        else:
            help_str += (
                "mouse right + vertical motion: translate camera along its z axis\n"
            )
            help_str += (
                "mouse right + horizontal motion: rotate camera along its z axis\n"
            )
            help_str += "mouse left + vertical motion: rotate camera along its x axis\n"
            help_str += (
                "mouse left + horizontal motion: rotate camera along its y axis\n"
            )
            help_str += "CTRL + mouse left + vertical motion: translate camera along its y axis\n"
            help_str += "CTRL + mouse left + horizontal motion: translate camera along its x axis\n"
            help_str += "SHIFT + mouse left + vertical motion: change the camera field of view\n"

        print(help_str)


class Viewer:
    def __init__(
        self,
        file_or_mesh,
        display_texture_map=True,
        width=640,
        height=480,
        display_fps=True,
        title=None,
        use_moderngl=False,
        light_directional=(0, 0, -0.5),
        light_ambient=0.5,
        background_color=(1, 1, 1),
        use_antialiasing=True,
        use_light=True,
        fps_exp_average_decay=0.1,
        horizontal_fov=60,
        video_pattern="deodr_viewer_recording{id}.avi",
        video_format="MJPG",
    ):
        self.title = title
        self.scene = differentiable_renderer.Scene3D(sigma=1)
        self.set_mesh(file_or_mesh)
        self.windowname = f"DEODR mesh viewer:{self.title}"

        self.width = width
        self.height = height
        self.display_fps = display_fps
        self.use_moderngl = use_moderngl
        self.use_antialiasing = use_antialiasing
        self.use_light = use_light
        self.fps_exp_average_decay = fps_exp_average_decay
        self.last_time = None
        self.horizontal_fov = horizontal_fov
        self.video_writer = None
        self.recording = False
        self.video_pattern = video_pattern
        self.video_format = video_format

        if display_texture_map:
            self.display_texture_map()

        self.set_background_color(background_color)
        self.set_light(light_directional, light_ambient)
        self.recenter_camera()

        if use_moderngl:
            self.setup_moderngl()
        else:
            self.offscreen_renderer = None

        self.register_keys()

    def set_light(self, light_directional, light_ambient):
        self.light_directional = np.array(light_directional)
        self.light_ambient = light_ambient
        self.scene.set_light(
            light_directional=self.light_directional, light_ambient=light_ambient
        )

    def setup_moderngl(self):
        import deodr.opengl.moderngl

        self.offscreen_renderer = deodr.opengl.moderngl.OffscreenRenderer()
        self.scene.mesh.compute_vertex_normals()
        self.offscreen_renderer.set_scene(self.scene)

    def set_background_color(self, background_color):
        self.scene.set_background_color(background_color)

    def display_texture_map(self):
        if self.mesh.textured:
            ax = plt.subplot(111)
            self.mesh.plot_uv_map(ax)

    def set_mesh(self, file_or_mesh):
        if isinstance(file_or_mesh, str):
            if self.title is None:
                self.title = file_or_mesh
            mesh_trimesh = trimesh.load(file_or_mesh)
            self.mesh = ColoredTriMesh.from_trimesh(mesh_trimesh)
        elif isinstance(file_or_mesh, trimesh.base.Trimesh):
            mesh_trimesh = file_or_mesh
            self.mesh = ColoredTriMesh.from_trimesh(mesh_trimesh)
            if self.title is None:
                self.title = "unknown"
        elif isinstance(file_or_mesh, ColoredTriMesh):
            self.mesh = file_or_mesh
            if self.title is None:
                self.title = "unknown"
        else:
            raise (
                TypeError(
                    f"unknown type {type(file_or_mesh)} for input obj_file_or_trimesh,"
                    " can be string or trimesh.base.Trimesh"
                )
            )
        self.object_center = 0.5 * (
            self.mesh.vertices.max(axis=0) + self.mesh.vertices.min(axis=0)
        )
        self.object_radius = np.max(
            self.mesh.vertices.max(axis=0) - self.mesh.vertices.min(axis=0)
        )
        self.scene.set_mesh(self.mesh)

    def recenter_camera(self):
        camera_center = self.object_center + np.array([0, 0, 3]) * self.object_radius
        rotation = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        translation = -rotation.T.dot(camera_center)
        extrinsic = np.column_stack((rotation, translation))
        focal = 0.5 * self.width / np.tan(0.5 * self.horizontal_fov * np.pi / 180)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )

        distortion = [0, 0, 0, 0, 0]
        self.camera = differentiable_renderer.Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )

        self.interactor = Interactor(
            camera=self.camera,
            object_center=self.object_center,
            z_translation_speed=0.01 * self.object_radius,
            xy_translation_speed=3e-4,
        )

    def start(self, print_help=True, loop=True):
        """Open the window and start the loop if loop true."""
        if print_help:
            self.print_help()
        self.fps = 0
        cv2.namedWindow(self.windowname, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(self.windowname, self.width, self.height)
        cv2.setMouseCallback(self.windowname, self.interactor.mouse_callback)
        if loop:
            while cv2.getWindowProperty(self.windowname, 0) >= 0:
                self.refresh()

    def update_fps(self):
        new_time = time.perf_counter()
        if self.last_time is None:
            self.fps = 0
        elif self.fps == 0:
            self.fps = 1 / (new_time - self.last_time)
        else:
            new_fps = 1 / (new_time - self.last_time)
            self.fps = (
                1 - self.fps_exp_average_decay
            ) * self.fps + self.fps_exp_average_decay * new_fps
        self.last_time = new_time

    def refresh(self):
        self.width, self.height = cv2.getWindowImageRect(self.windowname)[2:]
        self.resize_camera()

        if self.use_moderngl:
            image = self.offscreen_renderer.render(self.camera)
        else:
            image = (self.scene.render(self.interactor.camera) * 255).astype(np.uint8)

        bgr_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        if self.recording:
            self.video_writer.write(bgr_image.astype(np.uint8))

        self.update_fps()
        if self.recording:
            cv2.circle(
                bgr_image,
                (image.shape[1] - 20, image.shape[0] - 20),
                8,
                (0, 0, 255),
                cv2.FILLED,
            )
        if self.display_fps:
            self.print_fps(bgr_image, self.fps)

        cv2.imshow(self.windowname, bgr_image)

        key = cv2.waitKey(1)
        if key > 0:
            self.process_key(key)

    def resize_camera(self):
        ratio = self.width / self.camera.width

        intrinsic = np.array(
            [
                [self.camera.intrinsic[0, 0] * ratio, 0, self.width / 2],
                [0, self.camera.intrinsic[1, 1] * ratio, self.height / 2],
                [0, 0, 1],
            ]
        )
        self.camera.intrinsic = intrinsic
        self.camera.width = self.width
        self.camera.height = self.height

    def print_fps(self, image, fps):
        font = cv2.FONT_HERSHEY_SIMPLEX
        bottom_left_corner_of_text = (20, image.shape[0] - 20)
        font_scale = 1
        font_color = (0, 0, 255)
        thickness = 2
        cv2.putText(
            image,
            "fps:%0.1f" % fps,
            bottom_left_corner_of_text,
            font,
            font_scale,
            font_color,
            thickness,
        )

    def print_help(self):
        """Print the help message."""
        help_str = ""
        help_str += "-----------------\n"
        help_str += "DEODR Mesh Viewer\n"
        help_str += "-----------------\n"
        help_str += "Keys:\n"
        for key, func in self.keys_map.items():
            help_str += f"{key}: {func.__doc__}\n"
        print(help_str)
        self.interactor.print_help()

    def toggle_renderer(self):
        """Toggle the renderer between DEODR cpu rendering and moderngl."""
        self.use_moderngl = not (self.use_moderngl)
        print(f"use_moderngl = { self.use_moderngl}")

        if self.use_moderngl and self.offscreen_renderer is None:
            self.setup_moderngl()

    def toggle_perspective_texture_mapping(self):
        """Toggle between linear texture mapping and perspective correct texture mapping."""
        if self.use_moderngl:
            print("can only use perspective correct mapping  when using moderngl")
        else:
            self.scene.perspective_correct = not (self.scene.perspective_correct)
            print(f"perspective_correct = {self.scene.perspective_correct}")

    def toggle_lights(self):
        """Toggle between uniform lighting vs directional + ambient."""
        self.use_light = not (self.use_light)
        print(f"use_light = { self.use_light}")

        if self.use_light:
            if self.use_moderngl:
                self.offscreen_renderer.set_light(
                    light_directional=np.array(self.light_directional),
                    light_ambient=self.light_ambient,
                )
            else:
                self.scene.set_light(
                    light_directional=np.array(self.light_directional),
                    light_ambient=self.light_ambient,
                )
        else:
            if self.use_moderngl:
                self.offscreen_renderer.set_light(
                    light_directional=(0, 0, 0),
                    light_ambient=1.0,
                )
            else:
                self.scene.set_light(light_directional=None, light_ambient=1.0)

    def toggle_edge_overdraw_antialiasing(self):
        """Toggle edge overdraw anti-aliasing (DEODR rendering only)."""
        if self.use_moderngl:
            print("no anti-aliasing available when using moderngl")
        else:
            self.use_antialiasing = not (self.use_antialiasing)
            print(f"use_antialiasing = {self.use_antialiasing}")
            if self.use_antialiasing:
                self.scene.sigma = 1.0
            else:
                self.scene.sigma = 0.0

    def pickle_scene_and_cameras(self):
        """Save scene and camera in a pickle file."""
        filename = os.path.abspath("scene.pickle")
        # save scene and camera in pickle file
        with open(filename, "wb") as file:
            # dump information to the file
            pickle.dump(self.scene, file)
        print(f"saved scene in {filename}")

        filename = os.path.abspath("camera.pickle")
        print(f"save scene in {filename}")
        with open(filename, "wb") as file:
            # dump information to the file
            pickle.dump(self.camera, file)
        print(f"saved camera in {filename}")

    def toggle_interactor_mode(self):
        """Change the camera interactor mode."""
        self.interactor.toggle_mode()
        self.interactor.print_help()

    def toggle_video_recording(self):
        """Start and stop video recording."""
        if not self.recording:
            id = 0
            while os.path.exists(self.video_pattern.format(**dict(id=id))):
                id += 1
            filename = self.video_pattern.format(**dict(id=id))

            self.video_writer = cv2.VideoWriter(
                filename,
                cv2.VideoWriter_fourcc(*self.video_format),
                30,
                (self.width, self.height),
            )
            self.recording = True
        else:
            self.video_writer.release()
            self.recording = False

    def register_keys(self):
        self.keys_map = {}
        self.register_key("h", self.print_help)
        self.register_key("r", self.toggle_renderer)
        self.register_key("p", self.toggle_perspective_texture_mapping)
        self.register_key("l", self.toggle_lights)
        self.register_key("a", self.toggle_edge_overdraw_antialiasing)
        self.register_key("d", self.pickle_scene_and_cameras)
        self.register_key("s", self.toggle_video_recording)
        self.register_key("t", self.toggle_interactor_mode)

    def register_key(self, key, func):
        self.keys_map[key] = func

    def process_key(self, key):
        chr_key = chr(key)
        if chr_key in self.keys_map:
            self.keys_map[chr(key)]()
        else:
            print(f"no function registered for key {chr_key}")


def run():
    obj_file = os.path.join(deodr.data_path, "duck.obj")
    Viewer(obj_file, use_moderngl=False).start()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="mesh_viewer", usage="%(prog)s [options]")
    duck_file = os.path.join(deodr.data_path, "duck.obj")
    parser.add_argument("mesh_file", type=str, nargs="?", default=duck_file)
    args = parser.parse_args()
    mesh_file = args.mesh_file
    Viewer(mesh_file, use_moderngl=True).start()



"""Examples with 3D mesh rendering using various backend and comparison with deodr."""

import os

import deodr
from deodr import differentiable_renderer
from deodr.triangulated_mesh import ColoredTriMesh

import imageio

import matplotlib.pyplot as plt

import numpy as np

from scipy.spatial.transform import Rotation

import trimesh


def run(obj_file, width=640, height=480, display=True):
    example_rgb(
        obj_file, width=width, height=height, display=display, display_moderngl=True
    )


def default_scene(
    obj_file, width=640, height=480, use_distortion=True, integer_pixel_centers=True
):

    mesh_trimesh = trimesh.load(obj_file)

    mesh = ColoredTriMesh.from_trimesh(mesh_trimesh)

    # rot = Rotation.from_euler("xyz", [180, 0, 0], degrees=True).as_matrix()
    rot = Rotation.from_euler("xyz", [180, 0, 0], degrees=True).as_matrix()

    camera = differentiable_renderer.default_camera(
        width, height, 80, mesh.vertices, rot
    )
    if use_distortion:
        camera.distortion = np.array([-0.5, 0.5, 0, 0, 0])

    bg_color = np.array((0.8, 0.8, 0.8))
    scene = differentiable_renderer.Scene3D(integer_pixel_centers=integer_pixel_centers)
    light_ambient = 0
    light_directional = 0.3 * np.array([1, -1, 0])
    scene.set_light(light_directional=light_directional, light_ambient=light_ambient)
    scene.set_mesh(mesh)
    scene.set_background_color(bg_color)
    return scene, camera


def example_rgb(display=True, save_image=False, width=640, height=480):
    obj_file = os.path.join(deodr.data_path, "duck.obj")
    scene, camera = default_scene(obj_file, width=width, height=height)
    image = scene.render(camera)
    if save_image:
        image_file = os.path.abspath(os.path.join(deodr.data_path, "test/duck.png"))
        os.makedirs(os.path.dirname(image_file), exist_ok=True)
        image_uint8 = (image * 255).astype(np.uint8)
        imageio.imwrite(image_file, image_uint8)
    if display:
        plt.figure()
        plt.title("deodr rendering")
        plt.imshow(image)
    return image


def example_channels(display=True, save_image=False, width=640, height=480):
    obj_file = os.path.join(deodr.data_path, "duck.obj")
    scene, camera = default_scene(obj_file, width=width, height=height)

    def normalize(v):
        if v.ndim == 3 and v.shape[2] < 3:
            nv = np.zeros((v.shape[0], v.shape[1], 3))
            nv[:, :, : v.shape[2]] = v
        else:
            nv = v
        return (nv - nv.min()) / (nv.max() - nv.min())

    scene.sigma = 0

    channels = scene.render_deferred(camera)
    if display:
        plt.figure()
        for i, (name, v) in enumerate(channels.items()):
            ax = plt.subplot(2, 4, i + 1)
            ax.set_title(name)
            ax.imshow(normalize(v))

    if save_image:
        for name, v in channels.items():
            image_file = os.path.abspath(
                os.path.join(deodr.data_path, f"test/duck_{name}.png")
            )
            os.makedirs(os.path.dirname(image_file), exist_ok=True)
            image_uint8 = (normalize(v) * 255).astype(np.uint8)
            imageio.imwrite(image_file, image_uint8)


def example_pyrender(display=True, save_image=False, width=640, height=480):
    import deodr.opengl.pyrender

    obj_file = os.path.join(deodr.data_path, "duck.obj")
    scene, camera = default_scene(
        obj_file, use_distortion=False, width=width, height=height
    )
    scene.sigma = 0  # removing edge overdraw antialiasing
    image_no_antialiasing = scene.render(camera)
    image_pyrender, depth = deodr.opengl.pyrender.render(scene, camera)
    if display:
        plt.figure()
        ax = plt.subplot(1, 3, 1)
        ax.set_title("deodr no antialiasing")
        ax.imshow(image_no_antialiasing)

        ax = plt.subplot(1, 3, 2)
        ax.set_title("pyrender")
        ax.imshow(image_pyrender)

        ax = plt.subplot(1, 3, 3)
        ax.set_title("difference")
        ax.imshow(np.abs(image_no_antialiasing - image_pyrender.astype(np.float) / 255))


def example_moderngl(display=True, width=640, height=480):
    import deodr.opengl.moderngl

    obj_file = os.path.join(deodr.data_path, "duck.obj")
    for integer_pixel_centers in [True, False]:
        scene, camera = default_scene(
            obj_file,
            width=width,
            height=height,
            integer_pixel_centers=integer_pixel_centers,
        )
        scene.sigma = 0  # removing edge overdraw antialiasing
        # adding some perturbation to get better test
        camera.extrinsic[1, 2] = camera.extrinsic[1, 2] + 0.1
        camera.extrinsic[1, 1] = camera.extrinsic[1, 1] * 0.9

        image_no_antialiasing = scene.render(camera)
        moderngl_renderer = deodr.opengl.moderngl.OffscreenRenderer()
        moderngl_renderer.set_scene(scene)
        image_moderngl = moderngl_renderer.render(camera)
        diff = np.abs(
            image_no_antialiasing.astype(np.float) * 255
            - image_moderngl.astype(np.float)
        )
        if display:
            plt.figure()
            ax = plt.subplot(1, 3, 1)
            ax.set_title("deodr no antialiasing")
            ax.imshow(image_no_antialiasing)

            ax = plt.subplot(1, 3, 2)
            ax.set_title("moderngl")
            ax.imshow(image_moderngl)

            ax = plt.subplot(1, 3, 3)
            ax.set_title("difference")
            ax.imshow(np.clip(10 * diff / 255, 0, 1))
            plt.show()

    assert np.sum(np.any(np.abs(diff) > 15, axis=2)) <= 3


if __name__ == "__main__":
    example_moderngl(display=True)
    example_rgb(save_image=False)
    example_channels(save_image=False)
    example_pyrender()
    plt.show()
